"""
$Id: base.py 2466 2020-07-10 13:20:48Z pe $

EBAS Offline Masterdata base class

History:
V.1.1.0  2016-10-11  pe  initial version

"""

import pickle
import six

class OfflineMasterDataBase(object):
    """
    Base class for handling Offline Masterdata
    """
    # file name
    META_FILE = None  # To be set by client class
    META = {}
    INIT = False


    @classmethod
    def read_pickle_file(cls):
        """
        Reads the mastedata from the pickle file.
        Parameters:
            None
        Returns:
            None
        """
        _file = open(cls.META_FILE, 'rb')
        # Pickle format is different between python2 and 3
        # As long as we use python2 pickles (we need to, because the DB is
        # sybase and we have only a db module for python2), we will decode
        # the python2 pickles in python3
        if six.PY2:
            cls.META = pickle.load(_file)
        else:
            cls.META = pickle.load(_file, encoding='latin1')
            # TODO: use latin1 as long as pickles are written with py2,
            # then remove encoding (use utf-8)

    @classmethod
    def write_pickle_file(cls, dbh):
        """
        Recreates the pickle file from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        cls.from_db(dbh)
        _file = open(cls.META_FILE, 'wb')
        pickle.dump(cls.META, _file)

    @classmethod
    def from_db(cls, dbh):
        """
        Update the cache from the database.
        Parameters:
            dbh     database handle
        Returns:
            None
        """
        raise NotImplementedError
        



"""
atmos_phys unit conversion routines for gasses.
"""

import re
from nilutility.datatypes import DataObject
from .base import ConvertBaseFactor
from .elements import ELEMENTS

R = 8.3144621  # Gas constant

MIXING = 0b00
MASSCONC_ELEM = 0b10
MASSCONC_TOTAL = 0b11

class ConvertGasBase(ConvertBaseFactor):  # pylint: disable=W0223
    # W0223: Method ... is abstract in base class but is not overridden
    """
    Base class for gas conversions (mixing ratio / mass concentration).

    The gas conversion classes have their own base class. A lot of the logic
    is generically handled here. The final classes (conversion for single
    coponens) are only implementing by setting those class attributes.

    Attributes:
        MASS_CONC_ELEMENT: For parameters where only the mass of one element of
                           a component is used as mass concentartion unit.
                           E.g. sulfor_dioxide, ug S/m3
                           (-> MASS_CONC_ELEMENT = 'S')
        MOLAR_MASS: molar mass of the component or, in case of
                    MASS_CONC_ELEMENT, the molar mass of that element
        COMP_NAME: The EBAS component name, can be used to doublecheck molar 
                   mass (using the parameter synonyms, parsing the chemical
                   formular)
    TODO: check COMP_NAME
    TODO: in case of MASS_CONC_ELEMENT, an additional type of conversion can be
          defined: from/to mass concentration using the whole molecule
          E.g. in case of sulfor_dioxide, ug S/m3, additional conversion to
          ug/m3
    """

    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = None
    MOLAR_MASS_ELEMENT = None
    COMP_NAME = None

    def __init__(self, from_unit, to_unit, roundoffset=0, maxround=None,
                 temperature=273.15, pressure=101325):
        # pylint: disable=R0913
        # R0913: Too many arguments
        """
        Set up conversion object.
        Parameters:
            from_unit:
            to_unit: conversion from and to unit. Must be checked in
                     _set_parameters of the final class.
            roundoffset: rounding offset (usually +1 for import and -1 for
                         export)
            maxround: maximum rounding threshold, do no round coarser then
                      this should be used for export conversion only
                      (rounding expressed like in round function, 0=integer,
                      1=1digit after comma)
            temperature: temperature (standard conditions)
            pressure: pressure (standard conditions)
        For gas conversions, temperature and pressure (standard conditions for
        mass concentrations) are additionally needed.
        """
        # additional attributes for standard conditions.
        self.temperature = temperature
        self.pressure = pressure
        # call base class init
        super(ConvertGasBase, self).__init__(
            from_unit, to_unit, roundoffset=roundoffset, maxround=maxround)

    @classmethod
    def _parse_unit(cls, unit):
        """
        Parses a unit string
        Returns:
            tuple (prefix_factor, conversion_type)
            prefix_factor   factor which accounts for yhe metrix prefix
                            m=1E-3, u=1E-6, n=1E-9, p=1E-12
            conversion type bit field from/to mass concentartion / mixing ratio
        Raises Value error
        """
        prefixes = {
            None: 1,
            'm': 1.0E-3,
            'u': 1.0E-6,
            'n': 1.0E-9,
            'p': 1.0E-12,
        }
        parts_per = {
            'ppm': 1.0E-6,
            'ppmv': 1.0E-6,
            'ppb': 1.0E-9,
            'ppbv': 1.0E-9,
            'ppt': 1.0E-12,
            'pptv': 1.0E-12,
        }
        if cls.MASS_CONC_ELEMENT:
            reg = re.match(
                '^([munp])?g{}/m3$'.format((' ' + cls.MASS_CONC_ELEMENT)),
                unit)
            if reg:
                return prefixes[reg.group(1)], MASSCONC_ELEM
        reg = re.match(
            '^([munp])?g/m3$',
            unit)
        if reg:
            return prefixes[reg.group(1)], MASSCONC_TOTAL
        
        reg = re.match('^([munp])?mol/mol$', unit)
        if reg:
            return prefixes[reg.group(1)], MIXING
        if unit in parts_per:
            return parts_per[unit], MIXING
        raise ValueError("Unit '{}' cannot be parsed".format(unit))

    @property
    def from_mixing(self):
        """
        Conversion from mixing ratio?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b1100 == 0

    @property
    def from_massconc(self):
        """
        Conversion from mass concentration?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b1000 == 0b1000

    @property
    def from_massconc_elem(self):
        """
        Conversion from mass concentration based on one element?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b1100 == 0b1000

    @property
    def from_massconc_total(self):
        """
        Conversion from mass concentration based on the whole molecule?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b1100 == 0b1100

    @property
    def to_mixing(self):
        """
        Conversion to mixing ratio?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b11 == 0

    @property
    def to_massconc(self):
        """
        Conversion to mass concentration?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b10 == 0b10

    @property
    def to_massconc_elem(self):
        """
        Conversion to mass concentration based on one element?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b11 == 0b10

    @property
    def to_massconc_total(self):
        """
        Conversion to mass concentration based on the whole molecule?
        Returns:
            bool
        """
        return self.parameters.conv_type & 0b11 == 0b11

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        Raises ValueError when unit cannot be parsed
        """
        try:
            from_prefix, from_type = self._parse_unit(self.from_unit)
            to_prefix, to_type = self._parse_unit(self.to_unit)
        except ValueError:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))
        self.parameters = DataObject(factor=None)
        self.parameters.conv_type = (from_type << 2) + to_type
        if self.from_mixing and self.to_massconc:
            if self.to_massconc_elem:
                molar_mass = self.__class__.MOLAR_MASS_ELEMENT
            else:
                molar_mass = self.__class__.MOLAR_MASS_TOTAL
            if molar_mass is None:
                raise ValueError(
                    "Conversion to {} is not supported, missing molar "
                    "mass".format(self.to_unit))
            self.parameters.factor = (
                (molar_mass / (R*self.temperature/self.pressure)) *
                (from_prefix / to_prefix))
        elif self.from_massconc and self.to_mixing:
            if self.from_massconc_elem:
                molar_mass = self.__class__.MOLAR_MASS_ELEMENT
            else:
                molar_mass = self.__class__.MOLAR_MASS_TOTAL
            if molar_mass is None:
                raise ValueError(
                    "Conversion from {} is not supported, missing molar "
                    "mass".format(self.from_unit))
            self.parameters.factor = (
                ((R*self.temperature/self.pressure) / molar_mass) *
                (from_prefix / to_prefix))
        else:
            self.parameters.factor = from_prefix / to_prefix
            if self.from_massconc_elem and self.to_massconc_total:
                if (self.__class__.MOLAR_MASS_TOTAL is None or
                    self.__class__.MOLAR_MASS_ELEMENT is None):
                    raise ValueError(
                        "Conversion from {} to {} is not supported, missing "
                        "molar mass".format(self.from_unit, self.to_unit))
                self.parameters.factor *= (self.__class__.MOLAR_MASS_TOTAL /
                                           self.__class__.MOLAR_MASS_ELEMENT)
            elif self.from_massconc_total and self.to_massconc_elem:
                if (self.__class__.MOLAR_MASS_TOTAL is None or
                    self.__class__.MOLAR_MASS_ELEMENT is None):
                    raise ValueError(
                        "Conversion from {} to {} is not supported, missing "
                        "molar mass".format(self.from_unit, self.to_unit))
                self.parameters.factor *= (self.__class__.MOLAR_MASS_ELEMENT /
                                           self.__class__.MOLAR_MASS_TOTAL)

    def conversion_string(self, from_val=None, to_val=None):
        """
        Generates a conversion parameter string for the conversion.
        E.g. "from nmol/mol to ug/m3 273.15 K, 1013.25 hPa, using factor 1.234",
             "from nmol/mol to ug/m3 at 273.15 K, 1013.25 hPa, no conversion "
                 "factor used (no nonzero, valid values)"
        Parameters:
            from_val:
            to_val:
               If from_val and to_val are given, the values are included in the
               string (used to document conversion of uncertainty or detection
               limit values).
        Returns a string for documenting the conversion.
        """
        from_val = "{} ".format(from_val) if from_val is not None else ""
        to_val = "{} ".format(to_val) if to_val is not None else ""
        if self.rounding is None:
            convf = "no conversion factor (no nonzero, valid values)"
        else:
            convf = "conversion factor {:.{rdigits}f}".format(
                self.parameters.factor, rdigits=max(self.rounding+2, 0))

        if self.from_mixing and self.to_massconc:
            return ("from '{}{}' to '{}{}' at {} K, {} hPa, {}").format(
                from_val, self.from_unit, to_val, self.to_unit,
                self.temperature, self.pressure/100.0, convf)
        elif self.from_massconc and self.to_mixing:
            return ("from '{}{}' at {} K, {} hPa to '{}{}', {}").format(
                from_val, self.from_unit, self.temperature, self.pressure/100.0,
                to_val, self.to_unit, convf)
        return ("from '{}{}' to '{}{}', {}").format(
            from_val, self.from_unit, to_val, self.to_unit, convf)


class ConvertNOx(ConvertGasBase):
    """
    Conversion class for NOx (NO, NO2, NOx) conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug N/m3).
    """
    MASS_CONC_ELEMENT = 'N'
    MOLAR_MASS_ELEMENT = ELEMENTS['N']['MolarMass']


class ConvertCarbonMonoxide(ConvertGasBase):
    """
    Conversion class for CO:
    Mixing ratio (nmol/mol) / mmass concentration (mg/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        1 * ELEMENTS['C']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'carbon_monoxide'


class ConvertAmmonia(ConvertGasBase):
    """
    Conversion class for NH3
    Mixing ratio (nmol/mol) / mmass concentration (ug N/m3).
    """
    COMP_NAME = 'ammonia'
    MOLAR_MASS_TOTAL = (
        1 * ELEMENTS['N']['MolarMass'] +
        3 * ELEMENTS['H']['MolarMass'])
    MASS_CONC_ELEMENT = 'N'
    MOLAR_MASS_ELEMENT = ELEMENTS['N']['MolarMass']


class ConvertHydrochloricAcid(ConvertGasBase):
    """
    Conversion class for HCl
    Mixing ratio (nmol/mol) / mmass concentration (ug Cl/m3).
    """
    MASS_CONC_ELEMENT = 'Cl'
    MOLAR_MASS_ELEMENT = ELEMENTS['Cl']['MolarMass']


class ConvertNitricAcid(ConvertGasBase):
    """
    Conversion class for HNO3
    Mixing ratio (nmol/mol) / mmass concentration (ug N/m3).
    """
    MASS_CONC_ELEMENT = 'N'
    MOLAR_MASS_ELEMENT = ELEMENTS['N']['MolarMass']


class ConvertNitrousAcid(ConvertGasBase):
    """
    Conversion class for HNO3
    Mixing ratio (nmol/mol) / mmass concentration (ug N/m3).
    """
    MASS_CONC_ELEMENT = 'N'
    MOLAR_MASS_ELEMENT = ELEMENTS['N']['MolarMass']


class ConvertSulphurDioxide(ConvertGasBase):
    """
    Conversion class for NOx (NO, NO2, NOx) conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug N/m3).
    """
    MASS_CONC_ELEMENT = 'S'
    MOLAR_MASS_ELEMENT = ELEMENTS['S']['MolarMass']


class ConvertOzone(ConvertGasBase):
    """
    Conversion class for ozone conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = 3 * ELEMENTS['O']['MolarMass']
    COMP_NAME = 'ozone'


class ConvertEthanal(ConvertGasBase):
    """
    Conversion class for ethanal conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        4 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'ethanal'


class ConvertEthanol(ConvertGasBase):
    """
    Conversion class for ethanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'ethanol'


class ConvertMethanal(ConvertGasBase):
    """
    Conversion class for methanal conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        1 * ELEMENTS['C']['MolarMass'] +
        2 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'methanal'


class ConvertPropanone(ConvertGasBase):
    """
    Conversion class for propanone conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        3 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'propanone'

class ConvertPropanal(ConvertGasBase):
    """
    Conversion class for propanal conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        3 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'propanal'


class ConvertNPropanol(ConvertGasBase):
    """
    Conversion class for n-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        3 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'n-propanol'


class ConvertEthanedial(ConvertGasBase):
    """
    Conversion class for glyoxal conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        2 * ELEMENTS['H']['MolarMass'] +
        2 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'ethanedial'


class Convert2Oxopropanal(ConvertGasBase):
    """
    Conversion class for methylglyoxal conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        3 * ELEMENTS['C']['MolarMass'] +
        4 * ELEMENTS['H']['MolarMass'] +
        2 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = '2-oxopropanal'


class Convert2Propenal(ConvertGasBase):
    """
    Conversion class for 2-propenal conversions:
    Mixing ratio (pmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        3 * ELEMENTS['C']['MolarMass'] +
        4 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = '2-propenal'


class Convert3Buten2One(ConvertGasBase):
    """
    Conversion class for 3-buten-2-one conversions:
    Mixing ratio (pmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = '3-buten-2-one'


class ConvertButanone(ConvertGasBase):
    """
    Conversion class for butanone conversions:
    Mixing ratio (pmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'butanone'


class ConvertStyrene(ConvertGasBase):
    """
    Conversion class for styrene conversions:
    Mixing ratio (pmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        8 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'styrene'

class ConvertHexanal(ConvertGasBase):
    """
    Conversion class for haxanal conversions:
    Mixing ratio (pmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        6 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'hexanal'


class ConvertNButanal(ConvertGasBase):
    """
    Conversion class for n-butanal conversions:
    Mixing ratio (pmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'n-butanal'


class ConvertNaphthalene(ConvertGasBase):
    """
    Conversion class for naphthalene conversions:
    Mixing ratio (pmol/mol) / mmass concentration (ng/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'naphthalene'


class ConvertPentanal(ConvertGasBase):
    """
    Conversion class for pentanal conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'pentanal'


class Convert2Methylpropenal(ConvertGasBase):
    """
    Conversion class for 2-methylpropenal conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = '2-methylpropenal'


class ConvertBenzaldehyde(ConvertGasBase):
    """
    Conversion class for benzaldehyde conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        7 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'benzaldehyde'


class Convert2Propanol(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        3 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = '2-propanol'

class Convert123Trimethylbenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-2-3-trimethylbenzene'


class Convert124Trimethylbenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-2-4-trimethylbenzene'


class Convert135Trimethylbenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-3-5-trimethylbenzene'


class Convert1Ethyl3Methylbenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-ethyl-3-methylbenzene'


class Convert1Ethyl4Methylbenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-ethyl-4-methylbenzene'


class Convert3Carene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '3-carene'


class ConvertAcenaphthene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        12 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'acenaphthene'


class ConvertAcenaphthylene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        12 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'acenaphthylene'


class ConvertAnthracene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        14 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'anthracene'


class ConvertFluorene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        13 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'fluorene'


class ConvertAlphaPinene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'alpha-pinene'


class ConvertBenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        6 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'benzene'


class ConvertBetaPinene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'beta-pinene'


class ConvertCamphene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'camphene'


class ConvertEthylbenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        8 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'ethylbenzene'


class ConvertLimonene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'limonene'


class ConvertLinalool(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        18 * ELEMENTS['H']['MolarMass'] +
        ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'linalool'


class ConvertMPXylene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        8 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'm-p-xylene'


class ConvertMyrcene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'myrcene'


class ConvertNPropylbenzene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-propylbenzene'


class ConvertMethane(ConvertGasBase):
    """
    Conversion class for methane conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        ELEMENTS['C']['MolarMass'] +
        4 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'methane'


class ConvertEthane(ConvertGasBase):
    """
    Conversion class for ethane conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'ethane'


class ConvertPropane(ConvertGasBase):
    """
    Conversion class for propane conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        3 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'propane'


class ConvertNButane(ConvertGasBase):
    """
    Conversion class for n-butane conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-butane'


class ConvertNPentane(ConvertGasBase):
    """
   Conversion class for n-pentane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass']
    )
    COMP_NAME = 'n-pentane'

class ConvertNHexane(ConvertGasBase):
    """
   Conversion class for n-hexane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        6 * ELEMENTS['C']['MolarMass'] +
        14 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-hexane'

class ConvertNHeptane(ConvertGasBase):
    """
   Conversion class for n-heptane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        7 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-heptane'


class ConvertNOctane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        8 * ELEMENTS['C']['MolarMass'] +
        18 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-octane'


class ConvertNNonane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        20 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-nonane'


class ConvertNDecane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        22 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-decane'


class ConvertNUndecane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        11 * ELEMENTS['C']['MolarMass'] +
        24 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-undecane'


class ConvertNDodecane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        12 * ELEMENTS['C']['MolarMass'] +
        26 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-dodecane'


class ConvertNTridecane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        13 * ELEMENTS['C']['MolarMass'] +
        28 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-tridecane'


class ConvertNTetradecane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        14 * ELEMENTS['C']['MolarMass'] +
        30 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-tetradecane'


class ConvertNPentadecane(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        15 * ELEMENTS['C']['MolarMass'] +
        32 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-pentadecane'


class ConvertNHexadecane(ConvertGasBase):
    """
    Conversion class for n-hexadecane conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        16 * ELEMENTS['C']['MolarMass'] +
        34 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'n-hexadecane'


class Convert2Methylpropane(ConvertGasBase):
    """
    Conversion class for 2-methylpropane conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '2-methylpropane'


class ConvertOXylene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        8 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'o-xylene'


class ConvertPCymene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        14 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'p-cymene'


class ConvertTertButylbenzene(ConvertGasBase):
    """
    Conversion class for tert-butylbenzene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        14 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'tert-butylbenzene'


class Convert1Butene(ConvertGasBase):
    """
    Conversion class for 1-butene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-butene'


class ConvertTrans2Butene(ConvertGasBase):
    """
    Conversion class for trans-2-butene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'trans-2-butene'


class ConvertCis2Butene(ConvertGasBase):
    """
    Conversion class for cis-2-butene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'cis-2-butene'


class Convert2Methyl2Butene(ConvertGasBase):
    """
    Conversion class for 2-methyl-2-butene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '2-methyl-2-butene'


class Convert1Ethyl2Methylbenzene(ConvertGasBase):
    """
    Conversion class for 1-ethyl-2-methylbenzene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-ethyl-2-methylbenzene'


class ConvertChloroethene(ConvertGasBase):
    """
    Conversion class for chloroethene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        3 * ELEMENTS['H']['MolarMass'] +
                  1 * ELEMENTS['Cl']['MolarMass'])
    COMP_NAME = 'chloroethene'


class Convert13Butadiene(ConvertGasBase):
    """
    Conversion class for 1-3-butadiene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        4 * ELEMENTS['C']['MolarMass'] +
        6 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-3-butadiene'




class Convert2Methylbutane(ConvertGasBase):
    """
   Conversion class for 2-methylbutane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass']
    )
    COMP_NAME = '2-methylbutane'


class Convert1Pentene(ConvertGasBase):
    """
   Conversion class for 1-pentene conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass']
    )
    COMP_NAME = '1-pentene'



class ConvertTrans2Pentene(ConvertGasBase):
    """
   Conversion class for trans-2-pentene conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass']
    )
    COMP_NAME = 'trans-2-pentene'


class ConvertIsoprene(ConvertGasBase):
    """
   Conversion class for isoprene conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass']
    )
    COMP_NAME = 'isoprene'


class ConvertCis2Pentene(ConvertGasBase):
    """
   Conversion class for cis-2-pentene conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        5 * ELEMENTS['C']['MolarMass'] +
        10 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'cis-2-pentene'


class ConvertDichloromethane(ConvertGasBase):
    """
   Conversion class for dichloromethane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        1 * ELEMENTS['C']['MolarMass'] +
        2 * ELEMENTS['H']['MolarMass'] +
        2 * ELEMENTS['Cl']['MolarMass'])
    COMP_NAME = 'dichloromethane'


class Convert2Methylpentane(ConvertGasBase):
    """
   Conversion class for 2-methylpentane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        6 * ELEMENTS['C']['MolarMass'] +
        14 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '2-methylpentane'


class Convert3Methylpentane(ConvertGasBase):
    """
   Conversion class for 3-methylpentane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        6 * ELEMENTS['C']['MolarMass'] +
        14 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '3-methylpentane'


class Convert1Hexene(ConvertGasBase):
    """
   Conversion class for 1-hexene conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        6 * ELEMENTS['C']['MolarMass'] +
        12 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '1-hexene'


class ConvertTrichloroethane(ConvertGasBase):
    """
   Conversion class for trichloroethane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        3 * ELEMENTS['H']['MolarMass'] +
        3 * ELEMENTS['Cl']['MolarMass'])
    COMP_NAME = 'trichloroethane'


class Convert12Dichloroethane(ConvertGasBase):
    """
   Conversion class for 1-2-dichloroethane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        4 * ELEMENTS['H']['MolarMass'] +
        2 * ELEMENTS['Cl']['MolarMass'])
    COMP_NAME = '1-2-dichloroethane'


class Convert224Trimethylpentane(ConvertGasBase):
    """
   Conversion class for 2-2-4-trimethylpentane conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        8 * ELEMENTS['C']['MolarMass'] +
        18 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = '2-2-4-trimethylpentane'


class ConvertTrichloroethene(ConvertGasBase):
    """
   Conversion class for trichloroethene conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        1 * ELEMENTS['H']['MolarMass'] +
        3 * ELEMENTS['Cl']['MolarMass'])
    COMP_NAME = 'trichloroethene'


class ConvertTetrachloroethene(ConvertGasBase):
    """
   Conversion class for tetrachloroethene conversions:
    Mixing ratio (nmol/mol) / mass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        2 * ELEMENTS['C']['MolarMass'] +
        4 * ELEMENTS['Cl']['MolarMass'])
    COMP_NAME = 'tetrachloroethene'


class ConvertSabinene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'sabinene'


class ConvertTerpinolene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        16 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'terpinolene'


class ConvertToluene(ConvertGasBase):
    """
    Conversion class for 2-propanol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        7 * ELEMENTS['C']['MolarMass'] +
        8 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'toluene'


class Convert14Dichlorobenzene(ConvertGasBase):
    """
    Conversion class for 1-4-dichlorobenzene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        6 * ELEMENTS['C']['MolarMass'] +
        4 * ELEMENTS['H']['MolarMass'] +
        2 * ELEMENTS['Cl']['MolarMass'])
    COMP_NAME = '1-4-dichlorobenzene'


class ConvertEucalyptol(ConvertGasBase):
    """
    Conversion class for eucalyptol conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        10 * ELEMENTS['C']['MolarMass'] +
        18 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'eucalyptol'


class ConvertLongicyclene(ConvertGasBase):
    """
    Conversion class for longicyclene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        15 * ELEMENTS['C']['MolarMass'] +
        24 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'longicyclene'


class ConvertIsoLongifolene(ConvertGasBase):
    """
    Conversion class for iso-longifolene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        15 * ELEMENTS['C']['MolarMass'] +
        24 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'iso-longifolene'


class ConvertBetaCaryophyllene(ConvertGasBase):
    """
    Conversion class for beta-caryophyllene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        15 * ELEMENTS['C']['MolarMass'] +
        24 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'beta-caryophyllene'


class ConvertBetaFarnesene(ConvertGasBase):
    """
    Conversion class for beta-farnesene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        15 * ELEMENTS['C']['MolarMass'] +
        24 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'beta-farnesene'


class ConvertAlphaHumulene(ConvertGasBase):
    """
    Conversion class for alpha-humulene conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        15 * ELEMENTS['C']['MolarMass'] +
        24 * ELEMENTS['H']['MolarMass'])
    COMP_NAME = 'alpha-humulene'


class ConvertNopinone(ConvertGasBase):
    """
    Conversion class for nopinone conversions:
    Mixing ratio (nmol/mol) / mmass concentration (ug/m3).
    """
    MASS_CONC_ELEMENT = None
    MOLAR_MASS_TOTAL = (
        9 * ELEMENTS['C']['MolarMass'] +
        14 * ELEMENTS['H']['MolarMass'] +
        1 * ELEMENTS['O']['MolarMass'])
    COMP_NAME = 'nopinone'


def convf_volume_standard(temp1=273.15, pres1=101325, temp2=273.15,
                          pres2=101325):
    """
    Conversion factor for converting concentrations from one condition to
    another.
    Parameters:
        temp1: temperature to convert from
        pres1: pressure to convert from
        temp2: temperature to convert to
        pres2: pressure to convert to
    Returns:
        conversion factor
    Attention:
        all temperatures in K, all pressures in Pa (not hPa!!!)
    """
    return (temp1 * pres2) / (temp2 * pres1)



"""
$Id: base.py 2519 2020-10-30 16:13:09Z pe $

EBAS OPeNDAP base class

"""

from ..basefile.file import EbasFile, EbasJoinedFile
from ..base import EBAS_IOFORMAT_OPENDAP
\
EBAS_OPENDAP_REQUESTTYPE_DDS = 1
EBAS_OPENDAP_REQUESTTYPE_DAS = 2
EBAS_OPENDAP_REQUESTTYPE_DATA = 3

class EbasOPeNDAPBase(EbasJoinedFile):  # pylint: disable=R0901, W0223,
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Base class for all EBAS OPeNDAP classes.
    """

    # EbasOPeNDAP should use lazy read (do not read all data from DB while
    # setting up the opject). Depending on the request, only parts or even
    # nothing is needed.
    LAZY_READ = True
    IOFORMAT = EBAS_IOFORMAT_OPENDAP

    def __init__(self, access_log=None, *args, **kwargs):
        """
        Class initialization.
        """
        EbasJoinedFile.__init__(self, *args, **kwargs)
        self.strict_global = True   # OPeNDAP should use strict_global
        self.dataset = None   # pydap dataset object (destination for output)
        self.cdm_variables = {}
        self.access_log = access_log 

    def gen_filename(self, createfiles=False, destdir=None):
        # pylint: disable-msg=W0221
        # W0221: Arguments number differs from overridden method
        """
        Generates a file name according to NILU file name convention.
        Use the base class functionality but set the file name extension.
        Parameter:
            createfiles   write to physical file?
            destdir       destination directory path for files
        Return:
            None
        """
        EbasFile.gen_filename(self, createfiles=createfiles, destdir=destdir,
                              extension=".dods")



"""
$Id: list_helper.py 2227 2019-02-12 11:42:24Z pe $

Module for list helper functions.
This module is just a helper for dealing with common list manipulation tasks.
"""

def all_equal(iterable, ignore_none=True, single_value=False):
    """
    Highly efficient method for checking if all values are equal.
    (At least very fast if not all are equal)

    Parameters:
        iterable     iterable to be checked
        ignore_none  ignore None values (returns true if all non-None values are
                     equal)
        single_value if true, "all equal" is also triggered if there is just one
                     value (strict definition of "all")
    Returns:
        True if all the elements are equal to each other, else False
        Special cases:
            No values (length 0) return False
            Only None values and ignore_none=True returns True
            Only None values and ignore_none=False returns False
    """
    from itertools import groupby
    iterable = tuple(x for x in iterable if x is not None) if ignore_none \
        else iterable
    if not single_value and len(iterable) <= 1:
        # if single_value: "all equal" only if more then one value
        return False
    grp = groupby(iterable)
    return next(grp, False) and not next(grp, False)

def all_none(iterable):
    """
    Highly efficient method for checking if all values are None.
    (At least very fast if not all are None)

    Parameters:
        iterable     iterable to be checked
    Returns:
        True if all the elements are None, else False
        Special cases:
            No values (length 0) return False
    """
    from itertools import groupby
    grp = groupby(iterable)
    try:
        first = next(grp)
    except StopIteration:
        return False
    return first[0] is None and not next(grp, False)



"""
nilutility/float_helper.py
$Id: float_helper.py 1167 2016-01-19 09:57:07Z pe $

Module for float helper functions.
This module is just a helper for dealing with common float manipulation tasks.
"""

import sys
import math

def float_almostequal(flt1, flt2):
    """
    Compares float values and allow epsilon for representation errors.
    Epsilon does not cover computation and rounding errors that might have
    influenced either of the two values before.
    Parameters:
        flt1, flt2    float values
    Returns:
        True if nearly equal, else False
    """
    if flt1 == flt2:
        return True

    # get binary exponent of the floats to compare
    # IEEE 754: e.g. 1.0dec --> 0.1bin ** 2^1
    exp = math.frexp(flt1)[1]
    if exp != math.frexp(flt2)[1]:
        # the two floats differ if have different exponent
        return False

    # this is the IEEE 754 machine epsilon for binary exponent == 1
    # i.e. the value that increases the rightmost digit in the (binary) mantissa
    eps = sys.float_info.epsilon
    # now multiply with binary exponent of the target
    # machine eps is given for exponent == 1, so multiplying with exp gives a
    # variation in the second bit from right in the mantissa. this is wanted.
    eps *= math.pow(2, exp)
    # now, eps == number that would just increases the 2nd rightmost digit of
    # the mantissa in flt1 and flt2

    if abs(flt1 - flt2) < eps:
        return True
    return False



"""
atmos_phys unit conversion routines for water solutions.
"""

from nilutility import numeric
from nilutility.datatypes import DataObject
from .base import ConvertBaseFactor
from .elements import ELEMENTS

class ConvertSulphate(ConvertBaseFactor):
    """
    Conversion class for sulphate in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'mg S/l':
            self.parameters.factor = \
                ELEMENTS['S']['MolarMass'] / \
                (ELEMENTS['S']['MolarMass'] + 4*ELEMENTS['O']['MolarMass'])
        elif self.from_unit == 'mg S/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['S']['MolarMass'] + 4*ELEMENTS['O']['MolarMass']) / \
                ELEMENTS['S']['MolarMass']
        elif self.from_unit == 'mg S/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['S']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg S/l':
            self.parameters.factor = ELEMENTS['S']['MolarMass'] / 1000.0
        elif self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = \
                1000.0 / (ELEMENTS['S']['MolarMass'] +
                          4 * ELEMENTS['O']['MolarMass'])
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['S']['MolarMass'] + 4 * ELEMENTS['O']['MolarMass']) \
                / 1000
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertAmmonium(ConvertBaseFactor):
    """
    Conversion class for ammonium in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'mg N/l':
            self.parameters.factor = \
                ELEMENTS['N']['MolarMass'] / \
                (ELEMENTS['N']['MolarMass'] + 4*ELEMENTS['H']['MolarMass'])
        elif self.from_unit == 'mg N/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['N']['MolarMass'] + 4*ELEMENTS['H']['MolarMass']) / \
                ELEMENTS['N']['MolarMass']
        elif self.from_unit == 'mg N/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['N']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg N/l':
            self.parameters.factor = ELEMENTS['N']['MolarMass'] / 1000.0
        elif self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = \
                1000.0 / (ELEMENTS['N']['MolarMass'] +
                          4 * ELEMENTS['H']['MolarMass'])
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['N']['MolarMass'] + 4 * ELEMENTS['H']['MolarMass']) \
                / 1000
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertNitrate(ConvertBaseFactor):
    """
    Conversion class for ammonium in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'mg N/l':
            self.parameters.factor = \
                ELEMENTS['N']['MolarMass'] / \
                (ELEMENTS['N']['MolarMass'] + 3*ELEMENTS['O']['MolarMass'])
        elif self.from_unit == 'mg N/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['N']['MolarMass'] + 3*ELEMENTS['O']['MolarMass']) / \
                ELEMENTS['N']['MolarMass']
        elif self.from_unit == 'mg N/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['N']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg N/l':
            self.parameters.factor = ELEMENTS['N']['MolarMass'] / 1000.0
        elif self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = \
                1000.0 / (ELEMENTS['N']['MolarMass'] +
                          3 * ELEMENTS['O']['MolarMass'])
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['N']['MolarMass'] + 3 * ELEMENTS['O']['MolarMass']) \
                / 1000
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertBicarbonate(ConvertBaseFactor):
    """
    Conversion class for bicarbonate (HCO3-) in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = \
                1000.0 / (ELEMENTS['H']['MolarMass'] + \
                          ELEMENTS['C']['MolarMass'] + \
                          3 * ELEMENTS['O'])
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['H']['MolarMass'] + ELEMENTS['C']['MolarMass'] + \
                 3 * ELEMENTS['O']['MolarMass']) / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertFluoride(ConvertBaseFactor):
    """
    Conversion class for fluoride (F-) in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['F']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = ELEMENTS['F']['MolarMass'] / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertCalcium(ConvertBaseFactor):
    """
    Conversion class for calcium in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['Ca']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = ELEMENTS['Ca']['MolarMass'] / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertChloride(ConvertBaseFactor):
    """
    Conversion class for chloride in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['Cl']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = ELEMENTS['Cl']['MolarMass'] / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertMagnesium(ConvertBaseFactor):
    """
    Conversion class for magnesium in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['Mg']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = ELEMENTS['Mg']['MolarMass'] / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertPhosphate(ConvertBaseFactor):
    """
    Conversion class for phosphate (PO4) in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = \
                1000.0 / (ELEMENTS['P']['MolarMass'] + \
                          4 * ELEMENTS['O']['MolarMass'])
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = \
                (ELEMENTS['P']['MolarMass'] + 4 * ELEMENTS['O']['MolarMass']) \
                / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertPotassium(ConvertBaseFactor):
    """
    Conversion class for potassium in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['K']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = ELEMENTS['K']['MolarMass'] / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))

class ConvertSodium(ConvertBaseFactor):
    """
    Conversion class for potassium in precipitation
    """

    def _set_parameters(self):
        """
        Sets the conversion parameters.
        Called from base class init.
        Parameters:
            None
        Returns:
            None
        """
        self.parameters = DataObject(factor=None)
        if self.from_unit == 'mg/l' and self.to_unit == 'umol/l':
            self.parameters.factor = 1000.0 / ELEMENTS['Na']['MolarMass']
        elif self.from_unit == 'umol/l' and self.to_unit == 'mg/l':
            self.parameters.factor = ELEMENTS['Na']['MolarMass'] / 1000.0
        else:
            raise ValueError("Conversion from {} to {} is not supported".format(
                self.from_unit, self.to_unit))




"""
Module fileformats: support basic file formats
$Id: __init__.py 2631 2021-03-26 17:44:02Z pe $
 ascii_fixedwidth: fixed column ascii format
 NasaAmes1001:     NasaAmes1001 file I/O
 ...
"""
__version__ = 'V.1.10.00-next-devel'



"""
EANET precipitation chemistry file read
"""

import os
import re
import datetime
from fileformats.CsvUnicode import CsvUnicodeReader
from nilutility.datetime_helper import DatetimeInterval
from nilutility.numeric import digits_stat
from .base import EanetPrecipBase, EanetPrecipForm, EanetPrecipVariable, \
    EanetError, \
    from_address


class EanetPrecipRead(EanetPrecipBase):  # pylint: disable=R0902
    # R0902: Too many instance attributes
    """
    Class for reading EANET precip files
    """

    def __init__(self):
        """
        Initialize class.
        """
        EanetPrecipBase.__init__(self)
        self._rawdata = None
        self._merge_samples = []

    def read(self, filename):
        """
        Read from file.
        Parameters:
            filename    path and file name
        Returns:
            None
        Raises:
            IOError     in case of I/O issues
            EanetError  other raeding errors
        """
        self.logger.info('reading file %s', filename)
        self.file_name = os.path.basename(filename)
        self.parse_filename()
        self._read_file()

    def parse_filename(self):
        """
        Parses site code from filename, warnings if filename is not standard.
        Parameters:
            None
        Returns:
            None
        """
        reg = re.match(
            '([A-Z]{3}[0-9]{3})_([0-9]{4})_(wet_deposition).(csv)',
            self.file_name)
        if not reg:
            self.warning("Unconventional file name '%s'.",
                         self.file_name)
            reg = re.match('([A-Z]{3}[0-9]{3})_', self.file_name)
            if not reg:
                self.warning(
                    "Unable to parse site code from file name.")
                return
        self.site_code = reg.group(1)

    def _read_file(self):
        """
        Read the file content.
        Parameters:
            None
        Returns:
            None
        Raises: EanetError in case of errors
        """
        with open(self.file_name, 'rb') as fil:
            rdr = CsvUnicodeReader(fil)
            self._rawdata = [x for x in rdr]
        self._check_dimensions()
        self._read_metadata()
        self._check_form_headers()
        self._read_form_sample()
        self._read_form_data()

    def _check_dimensions(self):
        """
        Truncat empty lines, check number of columns.
        """
        # truncate empty lines at the end:
        chomp = 0
        while self._rawdata[-1] == [u'' for x in range(len(self._rawdata[-1]))]:
            del self._rawdata[-1]
            chomp += 1
        if chomp:
            self.warning("Ignored %s empty lines", chomp,
                         row=len(self._rawdata)-1, col=0)

        # check all lines have 252 columns:
        problems = [i for i, x in enumerate(self._rawdata) if len(x) != 252]
        if problems:
            txt = ', '.join(
                ["line {} [{} cols]".format(i+1, len(self._rawdata[i]))
                 for i in problems[:5]])
            if len(problems) > 5:
                txt += "..."
            self.error(
                "File contains %s lines with not exactly 252 columns: %s",
                len(problems), txt)
            raise  EanetError("Exiting because of previous errors")

    def _read_metadata(self):
        """
        Read metadata from the file. Check consistency of retundant metadata.
        """
        self.country_name = None
        self.site_name = None
        self.laboratory = None
        self.reporter = None
        self.funnel_diameter = None

        metadata_positions = [
            ['Country', 'country_name',
             ("B4", "D4"),
             ("AX4", "AZ4"),
             ("CH4", "CJ4"),
             ("DO4", "DQ4"),
             ("EQ3", "ER3"),
             ("FB3", "FC3"),
             ("GG3", "GH3"),
             ("HK3", "HL3")],
            ['Laboratory', 'laboratory',
             ("B6", "D6"),
             ("AX6", "AZ6"),
             ("CH6", "CJ6"),
             ("DO6", "DQ6")],
            ['Site', 'site_name',
             ("Q4", "S4"),
             ("BI4", "BK4"),
             ("CS4", "CU4"),
             ("DX4", "DY4"),
             ("EV3", "EW3"),
             ("FH3", "FI3"),
             ("GN3", "GO3"),
             ("HR3", "HS3")],
            ['Reporter', 'reporter',
             ("Q6", "T6"),
             ("BI6", "BL6"),
             ("CS6", "CV6"),
             ("DX6", "DZ6")],
            ['Funnel diameter', 'funnel_diameter',
             ("AG4", "AK4", "AM4"),
             ("BY4", "CC4", "CE4"),
             ("DF4", "DJ4", "DL4"),
             ("EH4", "EL4", "EN4")]]
        for meta in metadata_positions:
            for pos in meta[2:]:
                # lable:
                row, col = from_address(pos[0])
                if self._rawdata[row][col].strip() != meta[0]+' :':
                    self.error(
                        "expected '%s', found '%s'", meta[0]+' :',
                        self._rawdata[row][col], row=row, col=col)
                # value:
                target = meta[1]
                row, col = from_address(pos[1])
                if self.__dict__[target]:
                    if not self._rawdata[row][col]:
                        self.warning("Empty %s", meta[0], row=row, col=col)
                    if self.__dict__[target] != self._rawdata[row][col]:
                        self.error(
                            "Inkonsistent %s: '%s', found '%s' before",
                            meta[0], self._rawdata[row][col],
                            self.__dict__[target], row=row, col=col)
                elif self._rawdata[row][col]:
                    self.__dict__[target] = self._rawdata[row][col]
                # optional unit for funnel diameter:
                if meta[0] == 'Funnel diameter' and len(pos) == 3:
                    row, col = from_address(pos[2])
                    if self._rawdata[row][col].strip() != '[mm]':
                        self.error(
                            "expected '[mm]', found '%s'",
                            self._rawdata[row][col], row=row, col=col)
            if self.__dict__[target]:
                self.__dict__[target] = self.__dict__[target].strip()
        if self.funnel_diameter:
            try:
                self.funnel_diameter = int(self.funnel_diameter)
            except ValueError:
                self.error("funnel diameter '%s' is not numeric")
        if self.errors:
            raise  EanetError("Exiting because of previous errors")

    def _check_form_headers(self):
        """
        Check the forms layout
        """
        for formname, formdef in self.__class__.LAYOUT.items():
            row, col = from_address(formdef['header'])
            if self._rawdata[row][col] != formname:
                self.error(
                    "Expected '%s' but found '%s'",
                    formname, self._rawdata[row][col], row=row, col=col)
        if self.errors:
            raise EanetError("Exiting because of previous errors")

    def _read_form_sample(self):
        """
        Read sample numbers and sample times from all forms.
        """
        self._merge_samples = []
        for _, formdef in self.__class__.LAYOUT.items():
            row, col = from_address(formdef['start'])
            if formdef['sample_number']:
                sample_numbers = []
                if self._rawdata[row][col] != formdef['sample_number']:
                    self.error(
                        "Expected '%s' but found '%s'",
                        repr(formdef['sample_number']),
                        repr(self._rawdata[row][col]), row=row, col=col)
                else:
                    for i in range(row+3, len(self._rawdata)):
                        try:
                            sample_numbers.append(int(self._rawdata[i][col]))
                        except ValueError:
                            self.error(
                                "Non-integer sample number '%s'",
                                self._rawdata[i][col], row=i, col=col)
                if not self.sample_numbers:
                    self.sample_numbers = sample_numbers
                elif self.sample_numbers != sample_numbers:
                    self.error(
                        "Sample numbers sequence inconsistent (was different "
                        "before)", row=row, col=col)
                col += 1
            if formdef['sample_times']:
                sample_times = []
                check = (
                    ('Sampling period', 0, 0),
                    ('Start', 1, 0),
                    ('End', 1, 2),
                    ('Date', 2, 0),
                    ('Time', 2, 1),
                    ('Date', 2, 2),
                    ('Time', 2, 3))
                for chk in check:
                    if self._rawdata[row+chk[1]][col+chk[2]] != chk[0]:
                        self.error(
                            "Expected '%s', found '%s'", chk[0],
                            self._rawdata[row+chk[1]][col+chk[2]],
                            row=row+chk[1], col=col+chk[2])
                for i in range(row+3, len(self._rawdata)):
                    start = self._parse_datetime(i, col)
                    end = self._parse_datetime(i, col+2)
                    if start and end:
                        try:
                            intv = DatetimeInterval(start, end)
                        except (TypeError, ValueError) as excpt:
                            self.error(str(excpt), row=i, col=col)
                            sample_times.append(None)
                        else:
                            if sample_times and sample_times[-1] == intv:
                                # we assume this is an overflow sample
                                # take the sum of precipitation and the weighed
                                # arithmetic mean for concentrations and
                                # make one sample.
                                if i-row-3 not in self._merge_samples:
                                    # only write warning with first
                                    self.warning(
                                        'Duplicate sample times. '
                                        'Assuming overflow sample, will be '
                                        'merged to one sample.',
                                        row=i, col=col)
                                    self._merge_samples.append(i-row-3)
                                continue
                            sample_times.append(DatetimeInterval(start, end))
                            # check overlaps
                            if len(sample_times) > 1:
                                # look for last valid sample:
                                for samp in reversed(sample_times[:-1]):
                                    if samp:
                                        if samp[1] > sample_times[-1][0]:
                                            self.error(
                                                "Sample time overlap with "
                                                "previous valid sample: %s %s",
                                                samp, sample_times[-1],
                                                row=i, col=col)
                                        break
                    else:
                        sample_times.append(None)
                if not self.sample_times:
                    self.sample_times = sample_times
                elif self.sample_times != sample_times:
                    self.error(
                        "Sample times sequence inconsistent (was different "
                        "before)", row=row, col=col)
                col += 4
        if self.errors:
            raise EanetError("Exiting because of previous errors")

    def _read_form_data(self):
        """
        Read the data for all forms
        """

        def _cvt_flt(str_val, row, col):
            """
            Convert value to float or None.
            Parameters:
                str_val  string value
            Returns:
                float, None for missing
            Raises:
                ValueError in case of syntax errors
            """
            if str_val == '':
                return None
            try:
                return float(str_val)
            except ValueError as excpt:
                self.error(
                    "syntax error in value '{}': {}".format(
                        str_val, str(excpt)), row=row, col=col)

        def _cvt_int(str_val, row, col):
            """
            Convert value to int or None.
            Parameters:
                str_val  string value
            Returns:
                float, None for missing
            Raises:
                ValueError in case of syntax errors
            """
            if not str_val.strip():
                return None
            try:
                return int(str_val)
            except ValueError as excpt:
                self.error(
                    "syntax error in value '{}': {}".format(
                        str_val, str(excpt)), row=row, col=col)

        self.forms = []
        amount = None
        for formname, formdef in self.__class__.LAYOUT.items():
            form = EanetPrecipForm(self, len(self.forms), formname, formdef)
            self.forms.append(form)
            row, col = from_address(formdef['start'])
            if formdef['sample_number']:
                col += 1
            if formdef['sample_times']:
                col += 4
            for varname, vardef in formdef['vars'].items():
                if self._rawdata[row][col] == varname or \
                        (self._rawdata[row][col] == '' and \
                         varname.startswith('<EMPTY ')):
                    if vardef['flags']:
                        flags = [[_cvt_int(self._rawdata[i][j], i, j)
                                  for j in range(col+1, col+1+vardef['flags'])
                                  if self._rawdata[i][j].strip()]
                                 for i in range(row+3, len(self._rawdata))]
                    else:
                        flags = None
                    if varname.startswith("Date") or \
                            varname.startswith("Note"):
                        # use str without conversion
                        values = [self._rawdata[i][col]
                                  for i in range(row+3, len(self._rawdata))]
                    else:
                        values = [_cvt_flt(self._rawdata[i][col], i, col)
                                  for i in range(row+3, len(self._rawdata))]
                    form.variables.append(
                        EanetPrecipVariable(
                            varname, self._rawdata[row+2][col], values, flags))
                    if 'is_amount' in vardef and vardef['is_amount']:
                        # save the amount data, for later merging of samples
                        amount = list(values)
                else:
                    self.error(
                        "Expected variable name '%s', found '%s'",
                        '' if varname.startswith('<EMPTY ') else varname,
                        self._rawdata[row][col], row=row, col=col)
                col += 1 + vardef['flags']
        if self._merge_samples:
            # second loop to merge overflow samples
            for i in reversed(sorted(self._merge_samples)):
                for form in self.forms:
                    formdef = form.layout
                    for varname, vardef in formdef['vars'].items():
                        # seconf iteration: merge overflow samples if needed
                        var = form.var_by_name(varname)
                        if vardef['merge'] == 'mean':
                            if var.samples[i-1] is not None and \
                                    amount[i-1] and \
                                    var.samples[i] is not None and \
                                    amount[i]:
                                # get the rounding for all values:
                                mindig = digits_stat(var.samples)[2]
                                rnd = 0 if mindig is None \
                                    else -mindig if mindig < 0 \
                                    else -(mindig-1) if mindig > 0 \
                                    else 0
                                # calculate weighted mean and round
                                var.samples[i-1] = round(
                                    (var.samples[i] * amount[i] + \
                                     var.samples[i-1] * amount[i-1]) / \
                                     (amount[i] + amount[i-1]), rnd)
                                if var.flags:
                                    var.flags[i-1] += var.flags[i]
                            elif var.samples[i] is not None and \
                                    amount[i]:
                                var.samples[i-1] = var.samples[i]
                                if var.flags:
                                    var.flags[i-1] = var.flags[i]
                            # else, leave samples[i-1]
                        elif vardef['merge'] == 'sum':
                            var.samples[i-1] += var.samples[i]
                            if var.flags:
                                var.flags[i-1] += var.flags[i]
                        elif vardef['merge'] == 'concat':
                            var.samples[i-1] += ',' + var.samples[i]
                            if var.flags:
                                var.flags[i-1] += var.flags[i]
                        elif vardef['merge'] == 'first':
                            pass
                        else:
                            raise RuntimeError('unknown merge option {}'.format(
                                vardef['merge']))
                        del var.samples[i]
                        if var.flags:
                            del var.flags[i]
                amount[i-1] += amount[i]
        if self.errors:
            raise EanetError("Exiting because of previous errors")





"""
ebas/nasa_ames/ebas_xml.py
$Id: xml.py 1527 2017-02-15 18:36:36Z pe $

EBAS XML module

! Attention !
Basic XML functionality is implemented in the Xml class.
The ebas/ebas_xml (this) module
implements an interface to the ebas domain class (DB)

History:
V.1.0.0  2014-08-05  toh  initial version

"""

from .write import EbasXMLPartialWrite

class EbasXML(EbasXMLPartialWrite):  #pylint: disable-msg=R0901
    # R0901: Too many ancestors
    # This is a quite big partial class, therefore we ignore this warning
    """
    EBAS I/O XML object.
    This is a partial class distributed over many source files.
    """



"""
$Id: write.py 2519 2020-10-30 16:13:09Z pe $

EBAS OPeNDAP Module: output functionality
"""

import os
import sys
import itertools
from six import string_types
from six.moves import urllib
from .base import EbasOPeNDAPBase
from ..base import FLAGS_ALL
from ...ebasmetadata import EbasMetadata
from ebas.domain.basic_domain_logic.time_period import estimate_period_code, \
    period_code_iso8601_duration
from ..base import EBAS_IOFORMAT_OPENDAP
from .base import EBAS_OPENDAP_REQUESTTYPE_DAS, \
    EBAS_OPENDAP_REQUESTTYPE_DATA
try:
    from pydap.exceptions import ConstraintExpressionError
    from pydap.model import DatasetType, BaseType, GridType, StructureType, \
        SequenceType
    from pydap.lib import walk, get_var
    from pydap.responses.dods import DODSResponse
    import numpy as np
except ImportError:
    # pydap and numpy are NOT dependencies for ebas-io. Thus, ebas-io can
    # be used w/o this modules.
    # However the functionality is limited. An error message will be thrown
    # in the entry points of ebas opendap writer (methods write, make_dataset)
    pass

import datetime
from ...fileset.xmlwrap import xmlwrap_fileheader, xmlwrap_filetrailer
from nilutility.string_helper import list_joiner
from nilutility.datetime_helper import DatetimeISO8601Duration, DatetimeInterval

class EbasOPeNDAPPartialWrite(EbasOPeNDAPBase):
    """
    Conversion of EBAS data to OPeNDAP
    """

    def write(self, createfiles=False, destdir=None, xmlwrap=False,
              fileobj=None, flags=FLAGS_ALL, datadef='EBAS_1.1',
              metadata_options=0, suppress=0):
        """
        Writes the I/O object to a dap file (or stdout).
        Parameters:
            createfiles   write to file, else write to stdout
                          (special case for fileobj)
            destdir       destination directory path for files
            fileobj       write to filehandle
            flags         flag column options: FLAGS_ONE_OR_ALL (only allowed
                          value for OPeNDAP)
            datadef       ebas data definition (EBAS_1 or EBAS_1.1)
            metadata_options
                          options for output, bitfield:
                          currently only EBAS_IOMETADATA_OPTION_SETKEY
            suppress      suppress selected consolidations (bitfield):
                              SUPPRESS_SORT_VARIABLES
                              SUPPRESS_METADATA_OCCURRENCE
        Returns:
            None
        """
        self.check_requirements()
        fil = self.prepare_write(
            True, createfiles=createfiles, destdir=destdir, xmlwrap=xmlwrap,
            fileobj=fileobj, datadef=datadef, metadata_options=metadata_options,
            suppress=suppress)
        self.make_dataset(EBAS_OPENDAP_REQUESTTYPE_DATA, datadef,
                          metadata_options=metadata_options)

        resp = DODSResponse(self.dataset)
        for data in resp:
            fil.write(data)
        self.finish_write(createfiles, xmlwrap)
        self.write_indexdb()

    def make_dataset(self, requesttype, projection=None, datadef='EBAS_1.1',
                     metadata_options=None):
        """
        Build the pydap dataset object according to the (ebas domain) content of
        the obect. Sets the dataset attributes.
        Entry point for the pydap webservice.
        Parameters:
            requesttype   type of request for which the dataset should be built
                          EBAS_OPENDAP_REQUESTTYPE_DDS,
                          EBAS_OPENDAP_REQUESTTYPE_DAS
                          EBAS_OPENDAP_REQUESTTYPE_DATA
            datadef       EBAS metadata standard (to be used in the metadata
                          variables (json encoded))
            projection    pydap projection object (in order to constrain the
                          result dataset)
            metadata_obtions
                          options for output, bitfield:
                          currently only EBAS_IOMETADATA_OPTION_SETKEY
        Returns:
            None
        """
        self.check_requirements()
        self.prepare_write(False, datadef=datadef,
                           metadata_options=metadata_options)
        self._make_dataset(requesttype, projection, metadata_options)
        
    def _make_dataset(self, requesttype, projection, metadata_options):
        """
        Build the pydap dataset object according to the (ebas domain) content of
        the obect. Sets the dataset attributes.
        Entry point for the pydap webservice.
        Parameters:
            requesttype   type of request for which the dataset should be built
                          EBAS_OPENDAP_REQUESTTYPE_DDS,
                          EBAS_OPENDAP_REQUESTTYPE_DAS
                          EBAS_OPENDAP_REQUESTTYPE_DATA
            projection    pydap projection object (in order to constrain the
                          result dataset)
        Returns:
            None
        """
        set_data = True if requesttype == EBAS_OPENDAP_REQUESTTYPE_DATA \
                        else False
        set_attributes = True \
                         if requesttype in (EBAS_OPENDAP_REQUESTTYPE_DAS, \
                                            EBAS_OPENDAP_REQUESTTYPE_DATA) \
                         else False

        # set up the dataset
        self.dataset = DatasetType(name='Dataset')
        if set_attributes:
            # das or data: global attributes are needed
            self.add_global_attributes(metadata_options)
        # set up the time variables
        self._setup_dataset_time(set_data=set_data,
                                 set_attributes=set_attributes)
        self._setup_dataset_variables(set_attributes=set_attributes)
        # project:
        if projection:
            self.dataset = self._apply_projection(projection)
        if set_data:
            self._fill_data()

    def check_requirements(self):
        """
        Check requirements for opendap output.
        pydap and numpy are NOT dependencies for ebas-io generally.
        Thus, ebas-io can  be used w/o this modules.
        However the functionality is limited. An error message is thrown here
        and no output produced.
        # Info: make dataset is also a direct entry point (from the webservice)
        """
        try:
            import pydap  # @UnusedImport pylint: disable=W0612
            # W0612: Unused variable 'pydap'
            # Unused import: pydap
            # --> this line is just used to get a nice error message instead
            # of an exception in case of requirements not been installed
        except ImportError:
            self.error("The python pydap module is not installed. OPeNDAP "
                       "output is not available.")
            return
        try:
            import numpy  # @UnusedImport pylint: disable=W0612
            # W0612: Unused variable 'numpy'
            # Unused import: numpy
            # --> this line is just used to get a nice error message instead
            # of an exception in case of requirements not been installed
        except ImportError:
            self.error("The python numpy module is not installed. OPeNDAP "
                       "output is not available.")
            return

    def add_global_attributes(self, metadata_options):
        """
        Add global netCFD, CF and ACDD attributes.
        See generate_global_acdd_attributes for details.
        """
        self.dataset.attributes['NC_GLOBAL'] = {}
        # ACDD attributes
        for att in self.generate_global_acdd_attributes(metadata_options):
            self.dataset.attributes['NC_GLOBAL'][att[0]] = att[1]
        # EBAS attributes
        for att in self.generate_global_ebas_attributes():
            self.dataset.attributes['NC_GLOBAL'][att[0]] = att[1]
        # some DODS specific GAs:
        self.dataset.attributes['DODS_EXTRA'] = {}
        self.dataset.attributes['DODS_EXTRA']['unlimited_dimension'] = \
            "time"

    def _apply_projection(self, projection):
        """
        Applies the projection to the pydap dataset.
        Parameters:
            projection
        Return:
            new (projected) dataset
        (Copy/Paste from pydap DefaultHandler (v.3.2))
        """
        res = DatasetType(name=self.dataset.name,
                          attributes=self.dataset.attributes)

        # first collect all the variables
        for p in projection:
            target, template = res, self.dataset
            for i, (name, slice_) in enumerate(p):
                candidate = template[name]

                # add variable to target
                if isinstance(candidate, StructureType):
                    if name not in target.keys():
                        if i < len(p) - 1:
                            # if there are more children to add we need to clear
                            # the candidate so it has only explicitly added
                            # children; also, Grids are degenerated into Structures
                            if isinstance(candidate, GridType):
                                candidate = StructureType(
                                    candidate.name, candidate.attributes)
                            candidate._keys = []
                        target[name] = candidate
                    target, template = target[name], template[name]
                else:
                    target[name] = candidate

        # fix sequence data to include only variables that are in the sequence
        for seq in walk(res, SequenceType):
            seq.data = get_var(self.dataset, seq.id)[tuple(seq.keys())].data

        # apply slices
        for p in projection:
            target = res
            for name, slice_ in p:
                name = urllib.parse.unquote(name).decode('utf8')
                target, parent = target[name], target
                if slice_:
                    if name in self.cdm_variables:
                        self.cdm_variables[name]['slice'] = slice_
                    if isinstance(target, BaseType):
                        target.data = target[slice_]
                    elif isinstance(target, SequenceType):
                        parent[name] = target[slice_[0]]
                    elif isinstance(target, GridType):
                        parent[name] = target[slice_]
                    else:
                        raise ConstraintExpressionError("Invalid projection!")
        return res

    def _fill_data(self):
        """
        Fill the data after projection.
        """
        for var in walk(self.dataset, BaseType):
            if var.name in ['time', 'time_bounds', 'metadata_time',
                            'metadata_time_bounds']:
                # time variables are always filled, and have been correctly
                # sliced, nothing to do
                continue

            name = urllib.parse.unquote(var.name).decode('utf8')
            if name not in self.cdm_variables:
                # must be a dimension... those are filled with data and
                # sliced already
                continue
            # prepare variable flags according to dimensions
            odv = self.cdm_variables[name]
            if 'slice' in odv:
                dims = []
                indxs = []
                if (name.endswith('_qc') and \
                    len(odv['dimensions']) != len(odv['slice'])-2) or \
                   (not name.endswith('_qc') and \
                    len(odv['dimensions']) != len(odv['slice'])-1):
                    raise ValueError('wrong number of dimensions in projection')
                for i in range(len(odv['dimensions'])):
                    # iterate over n-1 dimensions (exclude time dim for now)
                    # and for _qc variables: exclude the flag dimension for now 
                    dims.append([(odv['dimensions'][i], k) for k in odv['dimextents'][i][odv['slice'][i]]])
                    indxs.append([(slice(j, j+1),) for j in range(len(odv['dimextents'][i][odv['slice'][i]]))])
            else:
                dims = [[tuple([odv['dimensions'][i], k]) for k in odv['dimextents'][i]] for i in range(len(odv['dimensions']))]
                indxs = [[slice(j, j+1) for j in range(len(odv['dimextents'][i]))] for i in range(len(odv['dimensions']))]

            # ebasmetadata: calculate max stringlen first:
            maxlen_ebasmeta = 1
            grid_ebasmeta = []
            if var.name.endswith('_ebasmetadata'):
                # slice the metadata_time?
                if 'slice' in odv:
                    t_slc = odv['slice'][-1]
                else:
                    t_slc = slice(0, None)
                for elem in itertools.product(*dims):
                    idx = odv['variable_dimensions'].index(elem)
                    var_index = odv['ebas_variables'][idx]
                    # generate all metadta in json encoding here, reuse
                    # when filling the data below.
                    # quoting:
                    # ebas_metadata_json returns unicode or str (if only 7bit)
                    dims_ebasmeta = [
                        mdi[2].ebas_metadata_json(var_index)
                        for mdi in self.metadata_intervals[t_slc]]
                    grid_ebasmeta.append(dims_ebasmeta)
                    maxlen_ebasmeta = max(maxlen_ebasmeta,
                                          max([len(x) for x in dims_ebasmeta]))
                shape = var.data.shape
                #shape = tuple([len(odv['dimextents'][i])
                #               for i in range(len(odv['dimensions']))] + \
                #               [len(self.metadata_intervals)])
                dtype = 'U{}'.format(maxlen_ebasmeta)
                var.data = np.empty(shape, dtype=dtype)

            # now fill the data according to the dims list
            index_perm = itertools.product(*indxs) # index permutations
            iter_grid_ebasmeta = iter(grid_ebasmeta)
            for elem in itertools.product(*dims):
                data_ind = next(index_perm)
                idx = odv['variable_dimensions'].index(elem)
                var_index = odv['ebas_variables'][idx]
                # slice the time?
                if 'slice' in odv:
                    t_slc = odv['slice'][-1]
                else:
                    t_slc = slice(0, None)
                if var.name.endswith('_qc'):
                    # prepare variable flags according to dimensions
                    flagdim = 3
                    # slice the flag dimension?
                    if 'slice' in odv:
                        f_slc = odv['slice'][-2]
                    else:
                        f_slc = slice(0, None)
                    fdim = len(([0] * flagdim)[f_slc])
                    tdim = len(self.sample_times[t_slc])
                    data = np.empty(shape=(fdim, tdim), dtype=np.int32)
                    for i in range(fdim):
                        lst = [flag[f_slc][i] if len(flag[f_slc]) >= i+1 else 0  for flag in self.variables[var_index].flags[t_slc]]
                        data[i] = np.array(lst, dtype=float)
                elif var.name.endswith('_ebasmetadata'):
                    # prepare variable metadata according to metadata_time dimension
                    dtype = 'U{}'.format(maxlen_ebasmeta)
                    dims_ebasmeta = next(iter_grid_ebasmeta)
                    data = np.array(dims_ebasmeta, dtype=dtype)
                else:
                    data = np.array(
                        self.variables[var_index].values_[t_slc],
                        # last dimension is slice is time
                        dtype=float)
                    if self.access_log:
                        # TODO: call siblings to log the access, 2 reasons:
                        # - right now, a joined file does not even know whether it should log accesss (access_log is not inherited when the join method creates a new parent object
                        # - single intervals need to be logged, not the overall interval
                        self.access_log.access_dataset(
                            self.get_meta_for_var(var_index, 'setkey'),
                            #self.variables[var_index].values_.lazies[0].hdi_obj.ds.DS_SETKEY,
                            # does not work for non compound lazy read objects!
                            DatetimeInterval(self.sample_times[t_slc][0][0],
                                             self.sample_times[t_slc][-1][-1]))

                target = var.data 
                for slc in data_ind:
                    target = target[slc][0]
                target[:] = data
                        
                    

        # self.cdm_variables =
        # 'grid_name' : {
        #         'ebas_variables' = [index, index, ...]
        #         'dimensions' = (name, name, ....)
        #         'dimextents' = ([value, value, ...], ...)
        #                          ... for each dimension
        #         'variable_dimensions': (((dimension, value), ...), ...)
        #                          ... for each variable (for each dimension)
        #         'slice': None,
        # }
#                 var_index = self.varnames.index(basename)
#                 flags = self.variables[var_index].flags[self.slices[var_index][0]]
#                 flags = [y + [0] * (flagdim-len(y)) for y in flags]
#                 var.data = np.array(flags, dtype=np.int32)
#             else:
#                 # prepare variable data according to time dimension
#                 var_index = self.varnames.index(urllib.unquote(var.name).decode('utf8'))
#                 # slice lazy read objects!!!
#                 ### this worked already...:
#                 ### data = np.array(self.variables[var_index].values_,
#                 ###                 dtype=float)
#                 ### var.data = data[self.slices[var_index]]
#                 var.data = np.array(
#                     self.variables[var_index].values_[self.slices[var_index][0]],
#                     # for now (1d): take the first dimension slice ...[0]
#                     dtype=float)

    def _setup_dataset_time(self, set_data=False, set_attributes=False):
        """
        Parameters:
            set_data          bool indicator if data should be included
            set_attributes    bool indicator if attributes should be included
        Returns:
            None
        """
        # create the time variable:
        zero = datetime.datetime(1900, 1, 1)
        if set_data:
            seq = (((intv[0]+(intv[1]-intv[0])/2)-zero).total_seconds()/86400.0
                   for intv in self.sample_times)
            data = np.fromiter(seq, dtype=float)
        else:
            data = np.empty((len(self.sample_times),), dtype=float)
        self.dataset['time'] = BaseType(
            name='time', data=data,
            shape=(len(self.sample_times),), type=float)
        if set_attributes:
            # das or data: time attributes are needed
            self.dataset['time'].attributes['standard_name'] = "time"
            self.dataset['time'].attributes['long_name'] = "time of measurement"
            self.dataset['time'].attributes['units'] = \
                "days since 1900-01-01 00:00:00"
            self.dataset['time'].attributes['axis'] = "T"
            self.dataset['time'].attributes['calendar'] = "gregorian"
            self.dataset['time'].attributes['bounds'] = "time_bnds"
            self.dataset['time'].attributes['cell_methods'] = "mean"

        # create time_bounds
        if set_data:
            seq = (((intv[0]-zero).total_seconds()/86400.0,
                    (intv[1]-zero).total_seconds()/86400.0)
                   for intv in self.sample_times)
            data = np.array(list(seq), dtype=float)
        else:
            data = np.empty((len(self.sample_times), 2), dtype=float)
        self.dataset['time_bounds'] = BaseType(
            name='time_bounds', data=data, shape=(len(self.sample_times), 2),
            type=float, dimensions=('time', 'bnds'))
        if set_attributes:
            # das or data: time_bounds attributes are needed
            self.dataset['time_bounds'].attributes['standard_name'] = "time"
            self.dataset['time_bounds'].attributes['long_name'] = "time bounds for measurement"
            self.dataset['time_bounds'].attributes['units'] = "days since 1900-01-01 00:00:00"
            self.dataset['time_bounds'].attributes['calendar'] = "gregorian"

        # Setup metadata_time
        if set_data:
            seq = (((intv[0]+(intv[1]-intv[0])/2)-zero).total_seconds()/86400.0
                   for intv in self.metadata_intervals)
            data = np.fromiter(seq, dtype=float)
        else:
            data = np.empty((len(self.metadata_intervals),), dtype=float)
        self.dataset['metadata_time'] = BaseType(
            name='metadata_time', data=data,
            shape=(len(self.metadata_intervals), ), type=float)
        self.dataset['metadata_time'].attributes['standard_name'] = "time"
        self.dataset['metadata_time'].attributes['long_name'] = \
            "time of ebas metadata intervals"
        self.dataset['metadata_time'].attributes['units'] = \
            "days since 1900-01-01 00:00:00"
        self.dataset['metadata_time'].attributes['axis'] = "T"
        self.dataset['metadata_time'].attributes['calendar'] = "gregorian"
        self.dataset['metadata_time'].attributes['bounds'] = \
            "metadata_time_bnds"
        self.dataset['metadata_time'].attributes['cell_methods'] = "mean"

        # Setup metadata_time_bounds
        if set_data:
            seq = (((intv[0]-zero).total_seconds()/86400.0,
                    (intv[1]-zero).total_seconds()/86400.0)
                   for intv in self.metadata_intervals)
            data = np.array(list(seq), dtype=float)
        else:
            data = np.empty((len(self.metadata_intervals), 2), dtype=float)
        data.shape = (len(self.metadata_intervals), 2)
        self.dataset['metadata_time_bounds'] = BaseType(
            name='metadata_time_bounds', data=data, shape=data.shape,
            type=float, dimensions=('metadata_time', 'bnds'))
        self.dataset['metadata_time_bounds'].attributes['standard_name'] = \
            "time"
        self.dataset['metadata_time_bounds'].attributes['long_name'] = \
            "time bounds for ebas metadata intervals"
        self.dataset['metadata_time_bounds'].attributes['units'] = \
            "days since 1900-01-01 00:00:00"
        self.dataset['metadata_time_bounds'].attributes['calendar'] = \
            "gregorian"

    def _setup_grid(self, gridname, set_attributes=False):
        """
        Setup a grid variable within the result dataset.
        No data should be added yet (just create the dataset frame).
        """
        if gridname.endswith('_ebasmetadata'):
            self._setup_grid_ebasmetadata(gridname, set_attributes=set_attributes)
            return
        odv = self.cdm_variables[gridname]
        flagdim = 3
        # TODO: select the right flag dim for the datatset
        # this was veeery inefficient: 
        # flagdim = max([len(y) for x in self.metadata_intervals
        #               for y in x[2].variables[i].flags])
        if gridname.endswith('qc'):
            shape = tuple([len(odv['dimextents'][i])
                           for i in range(len(odv['dimensions']))] + \
                          [flagdim] + [len(self.sample_times)])
            dim = odv['dimensions'] + ('flags',) + ('time', )
            data = np.empty(shape, dtype=np.int32)
            type = np.int32
        else:
            shape = tuple([len(odv['dimextents'][i])
                           for i in range(len(odv['dimensions']))] + \
                          [len(self.sample_times)])
            dim = odv['dimensions'] + ('time', )
            data = np.empty(shape, dtype=float)
            type = float
        grid = GridType(name=gridname)
        grid[gridname] = BaseType(name=gridname, data=data, shape=shape,
                                  type=float, dimensions=dim)

        for i in range(len(odv['dimensions'])):
            if isinstance(list(odv['dimextents'][i])[0], string_types):
                data = [x for x in odv['dimextents'][i]]
                maxlen = max([len(x) for x in data])
                dtype = 'U{}'.format(maxlen)
                shape = (len(data), )
                data = np.array(data, dtype=dtype)
                grid[dim[i]] = BaseType(name=dim[i], data=data, shape=shape,
                                        type=str, dimensions=(dim[i],))
            else:
                # int or float:
                data = [float(x) for x in odv['dimextents'][i]]
                shape = (len(data), )
                data = np.array(data, dtype=float)
                grid[dim[i]] = BaseType(name=dim[i], data=data, shape=shape,
                                        type=float, dimensions=(dim[i],))
        # DO NOT create a mapping - not 100% sure yet whether this is correct. 
        #if gridname.endswith('qc'):
        #    grid['flags'] = BaseType(name='flags',
        #                             data=np.array([i+1 for i in range(flagdim)]),
        #                             shape=(flagdim,),
        #                             type=float)
        
        grid['time'] = self.dataset['time'].__copy__()
        # make a shallow copy of the time variable. data is not copied
        # per se, but when one of the copies is projected, two different
        # data copies will exist. Thus all copies of time can be projected
        # independently: This did not work:
        # grid['time'] = self.dataset['time'] (if one grid or the base
        # variable was subsetted, the "copies" were affected too)
        
        # use the first variable (of all dimensions) for the attributes
        var_ind = odv['ebas_variables'][0] 
        cfname, cfunit = self.cfparams(
            self.get_meta_for_var(var_ind, 'regime'),
            self.get_meta_for_var(var_ind, 'matrix'),
            self.variables[var_ind].metadata.comp_name,
            self.get_meta_for_var(var_ind, 'statistics'),
            self.variables[var_ind].metadata.unit,
            self.get_meta_for_var(var_ind, 'datalevel'))
        if set_attributes and gridname.endswith('qc'):
            # das or data: variable attributes are needed
            if cfname:
                grid.attributes['standard_name'] = cfname + " status_flag"
            grid.attributes['missing_value'] = np.int32(0)
            grid.attributes['_FillValue'] = np.int32(0)
            grid.attributes['units'] = "1"
        elif set_attributes:
            if cfname:
                grid.attributes['standard_name'] = cfname
            grid.attributes['missing_value'] = np.float64(np.nan)
            grid.attributes['_FillValue'] = np.float64(np.nan)
            if cfunit:
                grid.attributes['units'] = cfunit
            else:
                grid.attributes['units'] = self.variables[var_ind].metadata.unit
            grid.attributes['ancillary_variables'] = \
                "{0}_qc {0}_ebasmetadata".format(gridname)
            cell_methods = {
                #'68.27% lower confidence bound': ,
                #'68.27% upper confidence bound': ,
                #'absolute error': ,
                #'accuracy': ,
                'arithmetic mean': 'mean',
                #'detection limit': 'det.lim.',
                #'expanded uncertainty 2sigma': 'ExpUnc2s',
                #'expanded uncertainty 3sigma': 'ExpUnc3s',
                #'geometric mean': 'gmean',
                'maximum': 'maximum',
                'median': 'median',
                'minimum': 'minimum',
                #'percentile:15.87': 'prec1587',
                #'percentile:84.13': 'perc8413',
                #'precision': 'precision',
                'stddev': 'standard_deviation',
                'uncertainty': 'uncertainty',
            }
            stat = self.get_meta_for_var(var_ind, 'statistics')
            if self.variables[var_ind].metadata.comp_name.startswith(
                    'precipitation_amount') and stat == 'arithmetic mean':
                # special case: precipitation amount is a sum 
                grid.attributes['cell_methods'] = "time: sum"
            elif stat in cell_methods:
                grid.attributes['cell_methods'] = "time: {}".format(
                    cell_methods[stat])
            else:
                grid.attributes['cell_methods'] = "time: {}".format(stat)
            grid.attributes['ebas_metadata'] = self.ebas_metadata_json(var_ind)

        self.dataset[gridname] = grid

    def _setup_grid_ebasmetadata(self, gridname, set_attributes=False):
        """
        Setup a grid variable within the result dataset.
        No data should be added yet (just create the dataset frame).
        """
        odv = self.cdm_variables[gridname]
        shape = tuple([len(odv['dimextents'][i])
                       for i in range(len(odv['dimensions']))] + \
                       [len(self.metadata_intervals)])
        dim = odv['dimensions'] + ('metadata_time', )
        dtype = 'U1'
        data = np.empty(shape, dtype=dtype)            
        grid = GridType(name=gridname)
        grid[gridname] = BaseType(name=gridname, data=data, shape=shape,
                                  type=str, dimensions=dim)

        for i in range(len(odv['dimensions'])):
            if isinstance(list(odv['dimextents'][i])[0], string_types):
                data = [x for x in odv['dimextents'][i]]
                maxlen = max([len(x) for x in data])
                dtype = 'U{}'.format(maxlen)
                shape = (len(data), )
                data = np.array(data, dtype=dtype)
                grid[dim[i]] = BaseType(name=dim[i], data=data, shape=shape,
                                        type=str, dimensions=(dim[i],))
            else:
                # int or float:
                data = [float(x) for x in odv['dimextents'][i]]
                shape = (len(data), )
                data = np.array(data, dtype=float)
                grid[dim[i]] = BaseType(name=dim[i], data=data, shape=shape,
                                        type=float, dimensions=(dim[i],))

        grid['metadata_time'] = self.dataset['metadata_time'].__copy__()
        # make a shallow copy of the time variable. data is not copied
        # per se, but when one of the copies is projected, two different
        # data copies will exist. Thus all copies of time can be projected
        # independently: This did not work:
        # grid['time'] = self.dataset['time'] (if one grid or the base
        # variable was subsetted, the "copies" were affected too)

        if set_attributes:
            grid.attributes['long_name'] = \
                "ebas metadata for different time intervals; json encoded"
        self.dataset[gridname] = grid

    def _setup_dataset_variables(self, set_attributes=False):
        """
        Parameters:
            requesttype   type of request for which the dataset should be built
                          EBAS_OPENDAP_REQUESTTYPE_DDS,
                          EBAS_OPENDAP_REQUESTTYPE_DAS
                          EBAS_OPENDAP_REQUESTTYPE_DATA
            projection    pydap projection object (in order to constrain the
                          result dataset)
        Returns:
            None
        """
        self.gen_cdm_variables()

        for gridname in self.cdm_variables.keys():
            self._setup_grid(gridname, set_attributes=set_attributes)



"""
ebas/io/xml/base.py
$Id: base.py 2519 2020-10-30 16:13:09Z pe $

EBAS XML base class

"""

from ..basefile.file import EbasFile
from ..base import EBAS_IOFORMAT_XML

class EbasXMLBase(EbasFile):  # pylint: disable=R0901, W0223,
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Base class for all EBAS XML classes.
    """

    IOFORMAT = EBAS_IOFORMAT_XML

    def __init__(self, result_id, *args, **kwargs):
        """
        Class initioalization.
        Parameters:
            result_id    increasing number (id for the resultset to be
                         generated) - needed for Xml object creation
        """
        EbasFile.__init__(self, *args, **kwargs)
        self.strict_global = True   # EBAS XML should use strict_global
        # add additional attributes:
        self.xml = None
        self.result_id = result_id

    def gen_filename(self, createfiles=False, destdir=None):
        # pylint: disable-msg=W0221
        # W0221: Arguments number differs from overridden method
        """
        Generates a file name according to NILU file name convention.
        Use the base class functionality but set the file name extension.
        Parameter:
            createfiles   write to physical file?
            destdir       destination directory path for files
        Return:
            None
        """
        EbasFile.gen_filename(self, createfiles=createfiles, destdir=destdir,
                              extension=".xml")



"""
ebas/nasa_ames/read/parse.py
$Id: parse.py 2585 2020-12-09 21:25:41Z pe $

parser for file input functionality for EBAS NASA Ames module
parses NasaAmes1001 object and builds NasaAems

History:
V.1.0.0  2013-06-22  pe  initial version

"""

import re
import datetime
from nilutility.ndecimal import Decimal
from nilutility.datatypes import DataObject
from nilutility.datetime_helper import datetime_round, DatetimeInterval
from nilutility.string_helper import list_splitter, ListSplitterNewlineError, \
        ListSplitterNullbyteError
from ebas.domain.masterdata.dl import EbasMasterDL
from ebas.domain.masterdata.pr import EbasMasterPR
from ..base import EbasNasaAmesReadError
from ...base import EBAS_IOFORMAT_NASA_AMES
from ...basefile.base import  EbasMetadataInvalid
from .parse_variables import NasaAmesPartialReadParserVariables
from .parse_ebasmetadata import NasaAmesPartialReadParserEbasMetadata
from ....ebasmetadata import EbasMetadata, EbasMetadataError

class NasaAmesPartialReadParser(# pylint: disable=R0901, W0223
        NasaAmesPartialReadParserVariables,
        NasaAmesPartialReadParserEbasMetadata):
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    # W0223: ... is abstract in base class but is not overridden
    # this class is abstract as well (does not have any NotImplementedError)
    """
    Parser for Nasa Ames I/O object.
    This is a base class for NasaAmesPartialRead
    """

    def parse_nasa_ames_1001(self, skip_variables=None, ignore_parameter=False):
        """
        Converts the NasaAmes1001 object into an EBAS NasaAmes Object.
        Parameters:
            skip_variables    list of variable numbers to be skipped (variabzle
                              numbers start with 1 for the first variable after
                              end_time)
            ignore_parameters ignore errors related to paramters and units
                              this is needed when a file with non standard
                              vnames should be processed without importing it
                              into the ebas.domain.
        Returns:
            None
        Raises:
            EbasNasaAmesReadError
        """
        # read ahead through the NCOM line in order to get the Data definition
        self._get_file_format(
            self.nasa1001.data.NCOML,
            13 + self.nasa1001.data.NV+ 1 + len(self.nasa1001.data.SCOML))
        self.metadata.originator = self._parse_persons(self.nasa1001.data.ONAME,
                                                       2)
        self.logger.debug(u'originator={}'.format(self.metadata.originator))
        self._parse_org(self.nasa1001.data.ORG)
        self.metadata.submitter = self._parse_persons(self.nasa1001.data.SNAME,
                                                      4)
        self.logger.debug(u'submitter={}'.format(self.metadata.submitter))
        self._parse_projects()
        if self.nasa1001.data.IVOL != 1:
            self.error('IVOL must be 1', lnum=6)
        if self.nasa1001.data.NVOL != 1:
            self.error('NVOL must be 1', lnum=6)
        self.metadata.reference_date = self.nasa1001.data.DATE
        self.metadata.revdate = self.nasa1001.data.RDATE
        self._parse_xname(self.nasa1001.data.XNAME)
        if self.nasa1001.data.NV < 3:
            # end_time, value, numflag is minimum!
            self.error('NV can not be less than 3', lnum=10)
        for vnum in [k for k in range(len(self.nasa1001.data.VSCAL))
                     if self.nasa1001.data.VSCAL[k] != 1.0]:
            self.error('VSCAL[{}]: only 1 is allowed'.format(vnum), lnum=11)
        # parse ncoml before variables: that way FT_TYPE and CO_COMP_NAME is
        # known (needed for characteristics check)
        self._parse_ncoml(
            self.nasa1001.data.NCOML,
            13 + self.nasa1001.data.NV + 1 + len(self.nasa1001.data.SCOML))
        self._parse_variables(skip_variables=skip_variables,
                              ignore_parameter=ignore_parameter)
        # _parse_title depends on varioables!
        self._parse_title(
            self.nasa1001.data.NCOML[-1],
            13 + self.nasa1001.data.NV + 1 + len(self.nasa1001.data.SCOML) +
            len(self.nasa1001.data.NCOML))
        self._parse_scoml(self.nasa1001.data.SCOML, 13 + self.nasa1001.data.NV)
        # set method_analytical MUST happen before check_mandatory!
        self._set_method_analytical()
        # check mandatory must happen here, sets read_tmp.hea_error if critical
        # metadata are missing, so that data cannot be read
        self.check_mandatory_metadata()
        if self.internal.read_tmp.hea_error:
            self.logger.info("Skip reading data because of severe errors in "
                             "header")
            raise EbasNasaAmesReadError(
                "{} Errors, {} Warnings".format(self.errors,
                                                self.warnings))
        self._parse_data(
            self.nasa1001.data.DATA,
            13 + self.nasa1001.data.NV + 1 + len(self.nasa1001.data.SCOML) + 1 +
            len(self.nasa1001.data.NCOML))
        self.check_timespec_metadata()

    def check_mandatory_metadata(self):
        """
        Check mandatory metadata according to metadata definition.
        Parameters:
            None
        Returns:
            None
        """
        for elem in self.internal.ebasmetadata.metadata:
            if (elem['main'] & 4 or elem['main'] & 8) and \
               (elem['key'] not in self.metadata or \
               self.metadata[elem['key']] is None):
                self.error("mandatory metadata attribute '{}' missing"
                           .format(elem['tag']))
                if elem['main'] & 8:
                    self.internal.read_tmp.hea_error = True
            if elem['vname'] & 4 or elem['vname'] & 8:
                # check mandatory per variable (if elem[vname] is mandatory)
                for i in range(len(self.variables)):
                    file_vnum, _, file_metalnum = self.msgref_vnum_lnum(i)
                    if not self.get_meta_for_var(i, elem['key']):
                        errstr = ('variable[{}]: mandatory metadata attribute '
                                  "'{}' missing".format(file_vnum, elem['tag']))
                        self.error(errstr, lnum=file_metalnum)
                        if elem['vname'] & 8:
                            self.internal.read_tmp.hea_error = True

    def check_timespec_metadata(self):
        """
        Checks existing timesecific metadata and log errors if needed.
        Parameters:
            None
        Returns:
            None
        Checks:
            type TU/TI (will be corrected in some cases)

        INFO: check startdate/first sample start is done in _check_startdate()
        during parsing of the first data line.
        
        Set type code is not necessarily the same as Nasa Ames DX: Nasa Ames
        x values are float. If the (non 0) DX applies to all x-diffs (last digit
        may be +- 1 due to rounding), the DX is valid. However, in the EBAS
        domain starttimes are datetime (rounded to 0.1s). Here we can have a TI
        series, even if DX in the file is 0.
        Example: 0.000000, 0.041667, 0.083334 0.125000
                 0.083334 is an erroneous reporting
                 (=> 02:00:00.0576 rounded to 02:00:00.1 in the ebas domain)
                 Thus, the series will be TI
                 In Nasa Ames (float) it's just one digit difference in the last
                 digit, which is acceppted to allow for rounding.
        """
        diff = self.uniform_time()
        if diff == datetime.timedelta(0):
            if self.metadata.type != "TI":
                self.warning(
                    'Set type code {} is not consistent with timeseries data, '
                    'setting TI'.format(self.metadata.type))
                self.metadata.type = "TI"
        elif isinstance(diff, datetime.timedelta):
            if self.metadata.type != "TU":
                self.warning(
                    'Set type code {} is not consistent with timeseries data, '
                    'setting TU'.format(self.metadata.type))
                self.metadata.type = "TI"
        # Else (diff is None): Less than 2 samples, type is undefined.
        # We accept whatever value is given.
        # On file read, we don't want to generate error messages for this
        # (e.g. NRT timeseries are definitely TU, but there is always just
        # one sample in the file).

    def _get_file_format(self, ncoml, lnum):
        """
        Read through NasaAmes1001 NCOM block to find the essential metadata
        for classifying the file
            Data definition   mandatory, error when not set
            Data level        optional
        Parameters:
            ncoml    NCOM block as read by Nasa1001 (list of str)
            lnum     line number of NNCOML line (for error reporting)
        Return:
            None
        """
        for i, line in enumerate(ncoml):
            try:
                (tag, value) = self._list_splitter(line, lnum+i+1,
                                                   delimiter=":")
            except ValueError:
                pass
            else:
                if tag == 'Data definition':
                    if value == 'EBAS_1':
                        self.warning('Legacy data definition EBAS_1',
                                     lnum=lnum+i+1)
                    elif value != 'EBAS_1.1':
                        self.error('illegal data definition ' + value,
                                   lnum=lnum+i+1)
                        raise EbasNasaAmesReadError(
                            "{} Errors, {} Warnings".format(self.errors,
                                                            self.warnings))
                    self.metadata.datadef = value
                if tag == 'Data level':
                    conv_dl = {
                        '': None,
                        '2b': '1.5'
                    }
                    if value in conv_dl.keys():
                        self.warning(
                            "converting data level '{}' to '{}'".format(
                                value, conv_dl[value]))
                        value = conv_dl[value]
                    if value is not None:
                        master = EbasMasterDL()
                        try:
                            _ = master[value]
                        except KeyError:
                            self.error(
                                "illegal Data level {} (only {} allowed)"
                                .format(value,
                                        ', '.join([elem['DL_DATA_LEVEL']
                                                   for elem in master])))
                            self.metadata.datalevel = EbasMetadataInvalid(value)
                        else:
                            self.metadata.datalevel = value
        if not self.metadata.datadef:
            self.error('No Data definition specified', lnum=lnum+1)
            raise EbasNasaAmesReadError("{} Errors, {} Warnings".format(
                self.errors, self.warnings))
        # instantiate the ebas metadata object to be used for this fileformat:
        self.internal.ebasmetadata = EbasMetadata(
            self.metadata.datadef, data_format=EBAS_IOFORMAT_NASA_AMES,
            data_level=self.metadata.datalevel)

    def _parse_persons(self, name, lnum):
        """
        Parse person names (generic method for ONAME or SNAME).
        Syntax: "Last Name(s), First Name(s)" or
                "Last Name(s), First Name(s);Last Name(s), First Name(s);..."
        Parameters:
            name   oname/sname line read from file
            lnum   line number (for error reporting)
        Returns:
            List of PS data elements
        """
        labels = ('PS_LAST_NAME', 'PS_FIRST_NAME', 'PS_EMAIL', 'PS_ORG_NAME',
                  'PS_ORG_ACR', 'PS_ORG_UNIT', 'PS_ADDR_LINE1', 'PS_ADDR_LINE2',
                  'PS_ADDR_ZIP', 'PS_ADDR_CITY', 'PS_ADDR_COUNTRY', 'PS_ORCID')
        ret = []
        if name == 'MISSING INFORMATION':
            self.warning('MISSING INFORMATION in person metadata', lnum=lnum)
            return ret
        if self.metadata.datadef == 'EBAS_1':
            for name in self._list_splitter(name, lnum, delimiter=','):
                reg = re.match('^([^ ]+) (.*)$', name.strip())
                if reg:
                    last = reg.group(1).strip()
                    first = reg.group(2).strip()
                    if last == '' and first == '':
                        self.warning("Ignoring empty Person name ('{}')".
                                     format(name), lnum=lnum)
                    else:
                        ret.append(DataObject(
                            dict(zip(labels,
                                     [last if last != u'' else None,
                                      first if first != u'' else None] +
                                     [None] * (len(labels)-2)))))
                else:
                    self.error('Person name syntax error', lnum=lnum)
        else: # default (EBAS_1.1)
            for name in self._list_splitter(name, lnum, delimiter=';'):
                try:
                    (last, first) = self._list_splitter(name.strip(), lnum,
                                                        delimiter=',')
                except ValueError:
                    self.error('Person name syntax error', lnum=lnum)
                else:
                    last = last.strip()
                    first = first.strip()
                    if last == '' and first == '':
                        self.warning("Ignoring empty Person name ('{}')".
                                     format(name), lnum=lnum)
                    else:
                        ret.append(DataObject(
                            dict(zip(labels,
                                     [last if last != u'' else None,
                                      first if first != u'' else None] +
                                     [None] * (len(labels)-2)))))
        if not ret:
            self.error('At least one person must be specified', lnum=lnum)
        return ret

    def _parse_org(self, org): # pylint: disable-msg=R0914
        # R0914 Too many local variables
        """
        Parse ORG line (Organization).
        Syntax:
            "code, name, acronym, unit, addr1, addr2, zip, city, country"
        Example:
            "NO01L, Norwegian Institute for Air Research, NILU,," +\
            "Instituttveien 18,, N-2027, Kjeller, Norway"
        Parameters:
            org   org line read from file (str)
        Returns:
            None
        """
        if self.metadata.datadef == 'EBAS_1':
            reg = re.match('^ *([^ ,]*), (.*)$', org)
            if not reg:
                self.error('Organization syntax error, should be: code, name, '
                           '...', lnum=3)
                return  # stop parsing this line
            (code, name) = reg.groups()
            (acron, unit, addr1, addr2, zip_, city, country) = (None, ) * 7
        else:  # default (EBAS_1.1)
            vals = [elem.strip() if elem.strip() != u'' else None
                    for elem in self._list_splitter(org, 3, delimiter=',')]
            try:
                (code, name, acron, unit, addr1, addr2, zip_, city, country) = \
                    vals
            except ValueError:
                self.error(
                    'Organization syntax error, should be: code, name, '
                    'acronym, unit, addr1, addr2, zip, city, country', lnum=3)
                return  # stop parsing this line
        # check OR_CODE syntax
        try:
            self.internal.ebasmetadata.parse_org_code(code)
        except EbasMetadataError as excpt:
            self.error("Organization code: " + str(excpt), lnum=3)
            return  # stop parsing this line

        reg = re.match(r"^([A-Z][A-Z])(\d\d)([LO])$", code)
        number = int(reg.group(2))
        type_ = reg.group(3)
        self.metadata.org = DataObject(
            OR_CODE=code, OR_TYPE=type_,
            OR_NUMBER=number, OR_NAME=name, OR_ACRONYM=acron, OR_UNIT=unit,
            OR_ADDR_LINE1=addr1, OR_ADDR_LINE2=addr2, OR_ADDR_ZIP=zip_,
            OR_ADDR_CITY=city, OR_ADDR_COUNTRY=country)
        self.logger.debug(u'org={}'.format(self.metadata.org))

    def _parse_projects(self):
        """
        Parse MNAME line and setup project list.
        """
        projstr = self.nasa1001.data.MNAME
        if self.metadata.datadef == 'EBAS_1':
            projstr = re.sub(r'\(insert\)', '', projstr)
        elif re.search(r'\(insert\)', projstr):
            self.warning(
                "Use of '(insert)' for projects is deprecated in EBAS_1.1, "
                "will be ignored", lnum=5)
            projstr = re.sub(r'\(insert\)', '', projstr)
        self.metadata.projects = self._list_splitter(projstr, 5)
        if len(self.metadata.projects) < 1:
            self.error('At least one project must be specified', lnum=5)
        master = EbasMasterPR()
        for proj in self.metadata.projects:
            try:
                master[proj]
            except KeyError:
                self.error("invalid framework acronym '{}'".format(proj),
                           lnum=5)
        self.logger.debug(u'projects={}'.format(self.metadata.projects))

    def _parse_xname(self, xname):
        """
        Parse XNAME line.
        Syntax:
            "days from file reference point" or
            "Days from the file reference point (start_time)"
        Parameters:
            xname   xname line read from file (str)
        Returns:
            None
        """
        if not re.match(\
           r'^(d|D)ays from (the )?file reference point( \(start_time\))?$', \
           xname):
            self.error("XNAME is not supported: '{}'".format(xname), lnum=9)
        self.logger.debug('xname: days from file reference point')

    def _parse_scoml(self, scoml, lnum):
        """
        Parse the complete SCOML block.
        SCOML is completely ignored and skipped currently.
        Parameters:
            scoml   scoml lines read from file (list of str)
            lnum          line number (NSCOML) (for error reporting)
        Returns:
            None
        """
        if len(scoml) > 0:
            self.warning('{} SCOML lines ignored'.format(len(scoml)),
                         lnum=lnum)

    def _parse_ncoml(self, ncoml, lnum):
        """
        Parse the complete NCOML block.
        Parameters:
            ncoml   ncoml lines read from file (list of str)
            lnum          line number (NNCOML) (for error reporting)
        Returns:
            None
        """
        # Component name and Instrument type needs to be pre parsed (needed for
        # any characteristics in NCOM!
        if len(ncoml) == 0:
            self.error('NNCOML: 0 not allowed, NCOM block is mandatory',
                       lnum=lnum)

        pre_read = ('Component', 'Instrument type')
        for i in range(len(ncoml)-1):
            # Last NCOM line should be EBAS TITLE line
            lst = self._list_splitter(ncoml[i], lnum+i+1, delimiter=":")
            if len(lst) == 2:
                # ignore lines which do not have tag: value syntax (error in
                # second loop)
                (tag, value) = self._list_splitter(ncoml[i], lnum+i+1,
                                                   delimiter=":")
                if tag in pre_read:
                    self._parse_ebasmetadata(tag, value, lnum + i + 1)

        for i in range(len(ncoml)):
            # loop through all but last line
            # Last NCOM line should be EBAS TITLE line
            # can only pe parsed after variables have been read
            # see _parse_title_()
            if i == len(ncoml) - 1 and ':' not in ncoml[i]:
                # if last line does not contain a colon: break and parse title
                # this is the normal case for the last nncom line
                break
            # if last line contains a colon: parse metadata first and
            # finally parse for title (raising an error that last line should
            # be title line)
            lst = self._list_splitter(ncoml[i], lnum+i+1, delimiter=":")
            if len(lst) <= 1:
                self.error(
                    'NCOM[{}]: syntax error: not "Tag: value" syntax (no colon '
                    'found which separates the metadata tag from the value)'
                    .format(i+1), lnum=lnum+i+1)
            elif len(lst) > 2:
                self.error(
                    'NCOM[{}]: syntax error: not "Tag: value" syntax (more '
                    'then one colon found which separates the metadata tag '
                    'from the value). Colons in metadata values need to be '
                    'escaped (quote the value with double quotes, e.g. '
                    'Comment: "text with :")'.format(i+1) , lnum=lnum+i+1)
            else:
                (tag, value) = self._list_splitter(ncoml[i], lnum+i+1,
                                                   delimiter=":")
                if tag in ('Data definition', 'Data level') or tag in pre_read:
                    # skip, this has been parsed in _get_file_format
                    # or in the previous loop
                    continue
                self._parse_ebasmetadata(tag, value, lnum + i + 1)

    def _parse_title(self, title, lnum):
        """
        Parse the title line (last NCOML line).
        Parameters:
            title   title line read from file (str)
            lnum    line number (NNCOML) (for error reporting)
        Returns:
            None
        """
        if not re.match(' *start((time)|(_time)|(_DOY)) ', title):
            self.warning(
                u"NCOM: last NCOM line should be title line: '{}'. Title line "
                u"needs to start with 'starttime ', 'start_time ' or "
                u"'start_DOY'".format(title), lnum=lnum)
            return   # stop parsing this line
        titles = self._list_splitter(title, lnum)

        if len(titles) != self.nasa1001.data.NV + 1:
            self.warning(
                u"NCOM: last NCOM line: Number of title elements must be equal "
                u"NV+1 (number of dependent variables specified in line 10) -- "
                u"ignoring titles. Info: NV={}, Number of title elememts={}, "
                u"Titles='{}'".format(self.nasa1001.data.NV, len(titles),
                                      title),
                lnum=lnum)
            return   # stop parsing this line

        if self.internal.read_tmp.hea_error:
            return
            # something critical was wrong with the variable metadata
        for i in range(len(self.variables)):
            self.variables[i].metadata.title = \
                                    titles[self.internal.read_tmp.var_num[i]+1]

    def _parse_data(self, data, lnum):
        """
        Converts the NasaAmes1001 data to EBAS NasaAmes:
        Sets objects's sample_times and variables.
        Parameters:
            data    NasaAmes1001 data array (list of lists, 1 for each variable)
            lnum    line number (of first data line) (for error reporting)
        """
        if not self.msg_condenser.is_reset():
            raise RuntimeError('Uninitialized MessageCondenser')
        self._parse_data_sample_times(data, lnum)
        flag = False
        vnum = 0
        for i in range(1, len(data)-1):
            if i in self.internal.read_tmp.var_skipped:
                # nothing needs to be done with flag. Value and flag variables
                # always need to be skipped at the same time.
                continue
            if flag:
                # data[i+1]: +1 because data includes independent var at [0]
                flags = list(data[i+1])  # make a copy
                dt_decimal = all([True if isinstance(x, Decimal) else False
                                  for x in flags if x is not None])
                if None in flags:
                    errstr = "variable[{}]: ".format(i+1) +\
                             "Missing value for flag column is not allowed."
                    for ln_ in [n for n, x in enumerate(flags)
                                if x is None]:
                        self.msg_condenser.add(
                            self.error, self.msg_condenser.MSG_ID_FLAG_MISSING,
                            i+1, errstr, lnum+ln_)
                    # ignore missing flag and set to 0 for further checks
                    subst = Decimal("0.0") if dt_decimal else 0.0
                    flags = [subst if x is None else x for x in flags]
                # check if value is 0.... and not negative or >= 1.0
                errstr = "variable[{}]: flag value must start with '0.'".format(
                    i+1)
                for ln_ in [n for n, x in enumerate(flags)
                            if x < 0.0 or x >= 1.0]:
                    self.msg_condenser.add(
                        self.error, self.msg_condenser.MSG_ID_FLAG_ZERODOT,
                        i+1, errstr, lnum+ln_)
                # flag conversion:
                #  1) convert all flag variables to Decimal
                #  2) '{:f}'.format() with a decimal prints always all digits
                #     (float does not! - only 6 digits after comma)
                if not dt_decimal:
                    flags = [Decimal(repr(f)) for f in flags]
                self.variables[vnum].flags = \
                    [[int((flagstr[k:k+3]+'00')[:3])
                      for k in range(0, len(flagstr), 3)
                      if int(flagstr[k:k+3]) != 0]
                     # special case: if flag in Nasa was Decimal and the flag
                     # string was '0' or '0.', the Decimal will be Decimal('0')
                     # which crashes ...split('.')[1]
                     for flagstr in ['{:f}'.format(numflag).split('.')[1]
                                     if repr(numflag) != "Decimal('0')"
                                     else "000"
                                     for numflag in flags]]
                # set all previous variables' flags if needed
                for j in reversed(range(vnum)):
                    if self.variables[j].flagcol:
                        break
                    # make a full copy of the flag list (flag correction in
                    # checks might change flags for a single variable):
                    self.variables[j].flags = \
                        [list(x) for x in self.variables[vnum].flags]
                vnum += 1
                flag = False
                # the next column is no flag column, because this was one
            else:
                # data[i+1]: +1 because data includes independent var at [0]
                self.variables[vnum].values_ = data[i+1]
                if self.variables[vnum].flagcol:
                    flag = True
                    # the next column will be a flag column
                else:
                    vnum += 1
        # write all condensed messages and reset condenser object:
        self.msg_condenser.deliver()

    def _parse_data_sample_times(self, data, lnum):
        """
        Sets objects's sample_times from the NasaAmes1001 data.
        Convert columns 0 and 1 to start and endtime, apply starttime and
        timeref, check for time interval overlaps.
        Parameters:
            data    NasaAmes1001 data array (list of lists, 1 for each variable)
            lnum    line number (of first data line) (for error reporting)
        """
        for i in range(len(data[0])):
            # start time is a float (checked already in NasaAmes1001)
            start = self.metadata.reference_date +\
                    self.internal.timeoffset +\
                    datetime.timedelta(days=float(data[0][i]))
            start = datetime_round(start, 1) # round to 1/10 second
            # make sure end time is not missing (this is valid in the
            # NasaAmes1001 file object, but not here in ebas)
            if  data[1][i] is None:
                self.msg_condenser.add(
                    self.error, self.msg_condenser.MSG_ID_TIME_MISSING,
                    None, "Missing value for end time is not allowed.",
                    lnum+i)
                self.internal.read_tmp.hea_error = True
                # strictly, this is not a header error, but needs to be
                # handled the same way (further parsing/checking of the
                # object needs to be skipped)
                end = None
            else:
                end = self.metadata.reference_date +\
                      self.internal.timeoffset +\
                      datetime.timedelta(days=float(data[1][i]))
                end = datetime_round(end, 1) # round to 1/10 second
            try:
                self.sample_times.append(DatetimeInterval(start, end))
            except ValueError:
                self.msg_condenser.add(
                    self.error, self.msg_condenser.MSG_ID_TIME_INVERT,
                    None, "starttime is not < endtime", lnum+i) 

    def _list_splitter(self, string, lnum, **kwargs):
        """
        Splits a string. Double quoted parts are untouched.
        Wrapper for nilutility.string_helper (add error reporting for
        parsing errors).
        Parameters:
            string        string to be splitted
            lnum          line number (for error reporting)
        Returns:
            [part,...] splitted parts as list
        """
        try:
            return list_splitter(string, **kwargs)
        except ListSplitterNewlineError:
            # seems to be an old issue: "newline inside string" seems to be
            # not thrown anymore in revcent versions of the csv module.
            # however on ratatoskr this is currently a problem
            # (status 2016-12-06, ubuntu 12.04, python 2.7.3)
            # this is an issue, we catch the _csv.Error exception in nilutility,
            # throw a ListSplitterNewlineError which is cought here.
            # TODO: may be deleted in th future
            # e.g. ubuntu 16.04, python 2.7.11, this is not an issue
            #      MacOS, python.org 2.7.10 also no issue
            self.error(
                'syntax error: newline inside quoted string constant is not '
                'allowed', lnum=lnum)
            return [string]
        except ListSplitterNullbyteError:
            self.error(
                'syntax error: NULL byte is not allowed', lnum=lnum)
            return [string]




"""
ebas/input_output/ebas_csv/ebas_csv.py
$Id: csv.py 1527 2017-02-15 18:36:36Z pe $

EBAS CSV module

! Attention !
Basic CSV functionality is implemented in the Csv class.
The ebas/ebas_csv (this) module
implements an interface to the ebas domain class (DB)

History:
V.1.0.0  2014-07-16  toh  initial version

"""

from .write import EbasCSVPartialWrite

class EbasCSV(EbasCSVPartialWrite):  # pylint: disable=R0901
    # R0901: Too many ancestors
    # This is a quite big class, therefore we ignore this warning
    """
    EBAS I/O CSV object.
    This is a partial class distributed over many source files.
    """



"""
$Id$

EBAS NetCDF module
"""

from .write import EbasNetCDFPartialWrite
from ..base import EBAS_IOFORMAT_NETCDF, EBAS_IOFORMAT_NETCDF1


class EbasNetcdf(EbasNetCDFPartialWrite):   #pylint: disable-msg=R0901
    # R0901: Too many ancestors
    # This is a quite big partial class, therefore we ignore this warning
    """
    EBAS I/O NetCDF.
    This is a partial class distributed over many source files.
    """
    IOFORMAT = EBAS_IOFORMAT_NETCDF

    def __init__(self, *args, **kwargs):
        """
        Initialize the object.
        NetCDF V2 should use strict_global.
        """
        super(EbasNetcdf, self).__init__(*args, **kwargs)
        self.strict_global = True

class EbasNetcdf1(EbasNetCDFPartialWrite):   #pylint: disable-msg=R0901
    # R0901: Too many ancestors
    # This is a quite big partial class, therefore we ignore this warning
    """
    EBAS I/O NetCDF Version 1. Should be discontinued in the future.
    This is a partial class distributed over many source files.
    """
    IOFORMAT = EBAS_IOFORMAT_NETCDF1

    def __init__(self, *args, **kwargs):
        """
        Initialize the object.
        NetCDF V1 should NOT use strict_global.
        """
        super(EbasNetcdf1, self).__init__(*args, **kwargs)
        self.strict_global = False



"""
$Id: sw.py 2466 2020-07-10 13:20:48Z pe $

EBAS Masterdata Class for Station WMO Region

This module implements the class EbasMasterSW.
"""

from .offline_masterdata import SWOfflineMasterData
from .base import EbasMasterBase

class EbasMasterSW(EbasMasterBase, SWOfflineMasterData):
    """
    Domain Class for station masterdata.
    Objects of this class do not represent entities, this class provides class
    methods for handling WMO region codes and checking them against master data.
    WMO region master data are retrieved from database or from offline storage
    when no database access is possible.
    """

    # read offline masterdata for SW (station WMO region code)
    # Those are fallback values, will be read from database as soon as possible.
    SWOfflineMasterData.read_pickle_file()

    def __init__(self, dbh=None):
        """
        Read st masterdata if dbh is provided.
        """
        SWOfflineMasterData.__init__(self)
        if not self.__class__.INIT and dbh:
            self.from_db(dbh)
